package de.kreawi.trainer;

import java.awt.CardLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Event;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;
import java.lang.Thread.UncaughtExceptionHandler;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.Observable;
import java.util.Observer;
import java.util.Properties;
import java.util.concurrent.CancellationException;
import java.util.concurrent.ExecutionException;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextPane;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;
import javax.swing.UIManager;
import javax.swing.border.EtchedBorder;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.jgoodies.forms.util.FormUtils;
import com.jgoodies.looks.HeaderStyle;
import com.jgoodies.looks.LookUtils;
import com.jgoodies.looks.Options;
import com.jgoodies.looks.plastic.PlasticLookAndFeel;
import com.jgoodies.uif_lite.component.Factory;
import com.jquizengine.questions.Question;
import com.jquizengine.questions.TitledQuestionGroup;
import com.jquizengine.questions.collection.QuestionPool;

import de.kreawi.trainer.TrainerModel.NewQuizPreparedListener;
import de.kreawi.trainer.db.QuestionAttributesModel.IcuChangedListener;
import de.kreawi.trainer.db.TrainerSettingsModel.SettingsChangedListener;
import de.kreawi.trainer.dialogs.EvaluationDialog;
import de.kreawi.trainer.dialogs.FxDialog;
import de.kreawi.trainer.dialogs.HelpDialog;
import de.kreawi.trainer.dialogs.KreawiInternalFrame;
import de.kreawi.trainer.dialogs.MainDialogFx;
import de.kreawi.trainer.dialogs.PreferencesDialog;
import de.kreawi.trainer.dialogs.backup.TrainerBackup;
import de.kreawi.trainer.dialogs.history.McHistoryDialog;
import de.kreawi.trainer.dialogs.history.TextHistoryDialog;
import de.kreawi.trainer.dialogs.quiz.QuizDialog;
import de.kreawi.trainer.dialogs.selection.ExamSelection;
import de.kreawi.trainer.dialogs.selection.MCCategorySelectionDialog;
import de.kreawi.trainer.dialogs.selection.MCICUSelectionDialog;
import de.kreawi.trainer.dialogs.selection.TextCategorySelectionDialog;
import de.kreawi.trainer.dialogs.selection.TextICUSelectionDialog;
import de.kreawi.trainer.exceptions.TrainerStorageException;
import de.kreawi.trainer.license.KreawiLicenseHandler;
import de.kreawi.trainer.license.RegistrationHandler;
import de.kreawi.trainer.onlinegame.net.OnlineGameCommunicator.ServerFailureCallback;
import de.kreawi.trainer.quiz.KreawiQuiz;
import de.kreawi.trainer.tutorial.TutorialHandler;
import de.kreawi.trainer.util.OSDetector;
import de.kreawi.trainer.util.TrainerHtmlUtil;
import de.kreawi.trainer.util.TrainerTextResources;
import de.kreawi.trainer.util.TrainerUtil;
import de.kreawi.trainer.util.swingx.CursorToolkit;
import de.kreawi.trainer.util.swingx.DesktopApi;
import de.kreawi.trainer.util.swingx.PleaseWaitDialog;
import de.kreawi.trainer.util.swingx.WindowToolkit;
import de.martinknauer.awt.Gradient.Direction;
import de.martinknauer.help.HelpRegistry;
import de.martinknauer.swing.GradientPanel;
import de.martinknauer.util.Util;
import de.martinknauer.util.config.Config;
import de.martinknauer.util.config.ConfigComponentHandler;
import edu.stanford.ejalbert.BrowserLauncher;
import javafx.application.Platform;
import net.roydesign.app.Application;

@SuppressWarnings("serial")
public class TrainerFrame extends JFrame {
  
  public static final String ACTION_SELECTED_KEY = "trainer.selected";
  
  public static final String ACTION_PROPOSE_EVALUATION_KEY = "trainer.proposeEvaluation";
  
  public static final String ACTION_CARD_BOX_LEVEL_KEY = "trainer.cardBoxLevel";
  
  public static final Dimension SMALL_DIALOG_SIZE = new Dimension(680, 560);
  
  public static final Dimension MEDIUM_DIALOG_SIZE = new Dimension(800, 630);
  
  public static final Dimension FRAME_SIZE = new Dimension(1008, 800);
  
  private TrainerMain main;
  
  private TrainerModel model;
  
  private Map<Actions, Action> actionMap;
  
  private CardLayout dialogCards;
  
  private JPanel mainPanel;
  
  private MainDialogFx mainDialog;
  
  private QuizDialog mcDialog;
  
  private EvaluationDialog evalDialog;
  
  private LatestVersionHandler latestVersionHandler;
  
  private TextHistoryDialog textHistoryDialog;
  
  private TextICUSelectionDialog textICUSelectionDialog;
  
  private ExamSelection textExamSelectionDialog;
  
  private TextCategorySelectionDialog textCategorySelectionDialog;
  
  private McHistoryDialog mcHistoryDialog;
  
  private MCICUSelectionDialog mcicuSelectionDialog;
  
  private ExamSelection mcExamSelectionDialog;
  
  private MCCategorySelectionDialog mcCategorySelectionDialog;
  
  private PleaseWaitDialog m_pleaseWaitDialog;

  private final StartupMessageQueue startupMessageQueue;

  private boolean m_closing = false;

  protected boolean fatalErrorOccurred = false;

  private static final Logger logger = LoggerFactory.getLogger(TrainerFrame.class); 
  
  /**
   * Configures the UI, then builds and opens the UI.
   */
  public TrainerFrame(TrainerMain main, TrainerModel model) {
    
    this.main = main;
    this.model = model;
    
    this.model.getQuestionAttributes().addIcuChangedListener(new IcuChangedListener() {
      @Override
      public void icuChanged() {
        updateIcuEnabledState();
      }
    });
    
    this.model.addNewQuizPreparedListener(new NewQuizPreparedListener() {
      @Override
      public void newQuizPrepared(KreawiQuiz quiz) {
        startNewQuiz();
      }
    });
    
    this.startupMessageQueue = initStartupMessageQueue();
    
  }
  
  /**
   * Creates and configures a frame, builds the menu bar, builds the content
   * and locates the frame on the screen. Called by TrainerMain.
   */
  void buildInterface() {
    
    if (KreawiGlobals.isDevMode() || KreawiGlobals.isBetaMode()) {
      logger.info("TrainerFrame.buildInterface: Method called!");
    }
    
    initActions();
    
    this.dialogCards = new CardLayout();
    
    if (Boolean.TRUE.equals(UIManager.getDefaults().get(KreawiInternalFrame.USE_GRADIENT_BACKGROUND_KEY))
        && !Boolean.TRUE.equals(UIManager.getDefaults().get(KreawiInternalFrame.USE_LOW_COLOR_DEPTH_BACKGROUND_KEY))) {
      this.mainPanel = new GradientPanel(this.dialogCards, Color.WHITE, Direction.VERTICAL, true);
    } else {
      this.mainPanel = new JPanel(this.dialogCards);
    }
    
    setContentPane(Factory.createStrippedScrollPane(this.mainPanel));
    
    if (KreawiGlobals.isDevMode() || KreawiGlobals.isBetaMode()) {
      logger.info("TrainerFrame.buildInterface: Adding MainDialog!");
    }
    this.mainDialog = new MainDialogFx(this.actionMap, this.model, this);
    this.mainPanel.add(this.mainDialog, MainDialogFx.class.toString());
    
    /* The next two lines are just for creating the splash screen */
    // this.mainPanel.setBackground(Color.WHITE);
    // this.mainPanel.setOpaque(true);
    
    addQuestionPanels();
    
    switchDialog(MainDialogFx.class.toString());
    
    setJMenuBar(buildMenuBar());
    
    setSizeAndLocation();
    
    final String suffix;
    if (KreawiGlobals.isDemo()) {
      suffix = TrainerTextResources.getString("APPLICATION_TITLE_SUFFIX_DEMO");
    } else if (KreawiGlobals.isBetaMode()) {
      suffix = TrainerTextResources.getString("APPLICATION_TITLE_SUFFIX_BETA");
    } else {
      suffix = "";
    }

    final String title = TrainerTextResources.getStringWithSubstitution("APPLICATION_TITLE_LONG", KreawiGlobals.CURRENT_YEAR)
        + (suffix.length() > 0 ? " " + suffix : "");

    setTitle(title);
    
    setIcon();
    
    setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
    this.addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e) {
        exit();
      }
    });
    
    updateIcuEnabledState();
    
    registerHelpItems();
    
    this.model.getSettings().addSettingChangedListener(new SettingsChangedListener() {
      @Override
      public void settingChanged(String key) {
        if (key.equals(TrainerSettingsConstants.KEY_MAX_CARD_BOX_LEVEL)
            || key.equals(TrainerSettingsConstants.KEY_DEFAULT_CARD_BOX_LEVEL)) {
          
          registerHelpItems();
          
        }
      }
    });
    
  }

  private StartupMessageQueue initStartupMessageQueue() {
    
    final StartupMessageQueue queue = new StartupMessageQueue();
    
    queue.addMessageTask(() -> possiblyAskForLicense());
    queue.addMessageTask(() -> possiblyAskForRegistration());
    queue.addMessageTask(() -> checkDataExistence());
    queue.addMessageTask(() -> checkForNewVersion());
    queue.addMessageTask(() -> possiblyShowTutorial());
    
    return queue;
  }

  public void setVisible(boolean b) {
    super.setVisible(b);
    
    Thread.setDefaultUncaughtExceptionHandler(new UncaughtExceptionHandler() {
      public void uncaughtException(Thread t, Throwable e) {
        logger.error("Nicht aufgefangener Anwendungsfehler!", e);
        if (!TrainerFrame.this.fatalErrorOccurred) {
          // don't show dialog twice!
          showFatalErrorDialogAndQuit();
          TrainerFrame.this.fatalErrorOccurred = true;
        }
      }
    });
    
    logger.info("TrainerFrame width: " + this.getWidth() + "; height: " + this.getHeight());
    
    startupMessageQueue.nextTask().run();
  }
  
  public void showFatalErrorDialogAndQuit() {
    final String title = TrainerTextResources.getString("APPLICATION_FATAL_ERROR_TITLE");
    final String message = TrainerTextResources.getString("APPLICATION_FATAL_ERROR_WITH_QUIT_TEXT");
    showMessageDialog(title, message, JOptionPane.ERROR_MESSAGE, new Runnable() {
      public void run() {
        exit();
      }
    });
  }
  
  public void showFatalErrorDialogWithoutQuit() {
    final String title = TrainerTextResources.getString("APPLICATION_FATAL_ERROR_TITLE");
    final String message = TrainerTextResources.getString("APPLICATION_FATAL_ERROR_WITHOUT_QUIT_TEXT");
    showMessageDialog(title, message, JOptionPane.ERROR_MESSAGE);
  }
  
  private void possiblyAskForLicense() {
    
    logger.info("Checking for valid license.");
    
    final Runnable nextTask = startupMessageQueue.nextTask();
    if (!KreawiGlobals.isDemo()) {
      nextTask.run();
      return;
    }

    if (KreawiLicenseHandler.getInstance().hasValidLicence()) {
      nextTask.run();
    } else {
      final Component parent = this;
      KreawiLicenseHandler.getInstance().askForLicenseKey(parent, nextTask);
    }
  }
  
  private void possiblyAskForRegistration() {
    logger.info("Checking for valid registration.");
    
    final Runnable nextTask = startupMessageQueue.nextTask();
    
    if (!KreawiLicenseHandler.getInstance().hasValidLicence())  {
      nextTask.run();
      return;
    }
    
    final Runnable exitTrainer = new Runnable() {
      public void run() {
        exit();
      }
    };
    
    final Component parent = this;
    SwingUtilities.invokeLater(new Runnable() {
      public void run() {
        RegistrationHandler.getInstance().registerLicense(parent, nextTask, exitTrainer);
      }
    });
  }
  
  private void checkDataExistence() {
    logger.info("Checking for full version data.");
    final Runnable nextTask = startupMessageQueue.nextTask();
    if (KreawiGlobals.dataExists()) {
      if (KreawiGlobals.isDemo() && RegistrationHandler.isFullVersionAllowed()) {
        askRestart();
      } else {
        nextTask.run();
      }
    } else if (RegistrationHandler.isFullVersionAllowed()) {
      final Component parent = this;
      SwingUtilities.invokeLater(new Runnable() {
        public void run() {
          final String message = TrainerTextResources.getString("DATA_DOWNLOAD_TEXT");
          final String title = TrainerTextResources.getString("DATA_DOWNLOAD_TITLE");
          String yesOption = TrainerTextResources.getString("DATA_DOWNLOAD_YES");
          String noOption = TrainerTextResources.getString("DATA_DOWNLOAD_NO");
          String[] options = new String[] { yesOption, noOption };
          final int answer = JOptionPane.showOptionDialog(parent, message, title, JOptionPane.YES_NO_OPTION,
              JOptionPane.QUESTION_MESSAGE, null, options, yesOption);
          if (answer == JOptionPane.YES_OPTION) {
            getLatestVersionHandler().showLatestVersionDialog();
          }
        }
      });
    } else {
      nextTask.run();
    }
  }
  
  private void askRestart() {
    final Component parent = this;
    
    SwingUtilities.invokeLater(new Runnable() {
      public void run() {
        
        final String message = TrainerTextResources.getString("APPLICATION_RESTART_TEXT");
        final String title = TrainerTextResources.getString("APPLICATION_RESTART_TITLE");
        String yesOption = TrainerTextResources.getString("APPLICATION_RESTART_YES");
        String noOption = TrainerTextResources.getString("APPLICATION_RESTART_NO");
        String[] options = new String[] { yesOption, noOption };
        final int answer = JOptionPane.showOptionDialog(parent, message, title, JOptionPane.YES_NO_OPTION,
            JOptionPane.QUESTION_MESSAGE, null, options, yesOption);
        if (answer == JOptionPane.YES_OPTION) {
          if (!restartApplication()) {
            final String errorMessage = TrainerTextResources.getString("APPLICATION_RESTART_ERROR");
            showMessageDialog(title, errorMessage, JOptionPane.WARNING_MESSAGE);
          }
        }
      }
    });
  }
  
  private void checkForNewVersion() {
    final Runnable nextTask = startupMessageQueue.nextTask();

    if (!RegistrationHandler.isRegistrationOk() || KreawiGlobals.isDemo() || !KreawiGlobals.dataExists()) {
      nextTask.run();
      return;
    }
    
    logger.info("Checking for updates.");
    
    final Thread thread = new Thread(new Runnable() {
      public void run() {
        getLatestVersionHandler().checkForNewVersionOnStartup(nextTask);
      }
    });
    thread.start();
  }
  
  protected void possiblyShowTutorial() {
    if (this.model.getSettings().getBooleanValue(TrainerSettingsConstants.KEY_SHOW_TUTORIAL_ON_STARTUP)) {
      TutorialHandler.showRandomTip(new Config(this.model.getSettings().getSettingProperties()), this);
    }
  }

  private LatestVersionHandler getLatestVersionHandler() {
    final TrainerFrame mainframe = this;
    if (latestVersionHandler == null) {
      latestVersionHandler = new LatestVersionHandler(mainframe, model, main);
    }
    return latestVersionHandler;
  }
  
  public boolean restartApplication() {
    try {
      if (OSDetector.getOs().isMac()) {
        final File kreawiApp = new File("/Applications/KreawiTrainer" + KreawiGlobals.CURRENT_YEAR + ".app");
        if (!kreawiApp.exists()) {
          throw new URISyntaxException(kreawiApp.getAbsolutePath(), "File doesn't exist!");
        }
        Runtime.getRuntime().exec(new String[] {"open", "-na", kreawiApp.getAbsolutePath()}, null, kreawiApp.getParentFile());
      } else {
        final String javaBin = System.getProperty("java.home") + File.separator + "bin" + File.separator + "java";
        final File currentJar = new File(TrainerMain.class.getProtectionDomain().getCodeSource().getLocation().toURI());
        
        /* is it a jar file? */
        if (!currentJar.getName().endsWith(".jar")) {
          logger.info("Application was not startet as jar file. Can't restart.");
          return false;
        }
        
        /* Build command: java -jar application.jar */
        final ArrayList<String> command = new ArrayList<String>();
        command.add(javaBin);
        command.add("-jar");
        command.add(currentJar.getPath());
        
        final ProcessBuilder builder = new ProcessBuilder(command);
        builder.start();
      }
      System.exit(0);
    } catch (URISyntaxException e) {
      logger.error("Error when trying to determine restart command. ", e);
      return false;
    } catch (IOException e) {
      logger.error("Error when trying to execute restart command. ", e);
      return false;
    }
    return true;
  }
  
  private void addQuestionPanels() {
    
    mcCategorySelectionDialog = new MCCategorySelectionDialog(this, this.model, this.actionMap,
        this.model.getMcQuestionsByCategories());
    this.mainPanel.add(mcCategorySelectionDialog, MCCategorySelectionDialog.class.toString());
    
    final QuestionPool mcQuestionsByExams = this.model.getMcQuestionsByExams();
    if (mcQuestionsByExams.getQuestionCount() > 0) {
      mcExamSelectionDialog = new ExamSelection(this, this.model, this.actionMap, mcQuestionsByExams);
      this.mainPanel.add(mcExamSelectionDialog, ExamSelection.class.toString());
    }
    
    mcicuSelectionDialog = new MCICUSelectionDialog(this, this.model, this.actionMap, this.model.getAllMcQuestions());
    this.mainPanel.add(mcicuSelectionDialog, MCICUSelectionDialog.class.toString());
    
    mcHistoryDialog = new McHistoryDialog(this, this.model, this.actionMap);
    this.mainPanel.add(mcHistoryDialog, McHistoryDialog.class.toString());
    
    /* generate question pool for text question selection dialogs */

    textCategorySelectionDialog = new TextCategorySelectionDialog(this, this.model, this.actionMap,
        this.model.getTextQuestionsByCategories());
    this.mainPanel.add(textCategorySelectionDialog, TextCategorySelectionDialog.class.toString());
    
    final QuestionPool textQuestionsByExams = this.model.getTextQuestionsByExams();
    if (textQuestionsByExams.getQuestionCount() > 0) {
      textExamSelectionDialog = new ExamSelection(this, this.model, this.actionMap, textQuestionsByExams);
      this.mainPanel.add(textExamSelectionDialog, ExamSelection.class.toString() + Question.Type.TEXT);
    }
    
    textICUSelectionDialog = new TextICUSelectionDialog(this, this.model, this.actionMap,
        this.model.getAllTextQuestions());
    this.mainPanel.add(textICUSelectionDialog, TextICUSelectionDialog.class.toString());
    
    textHistoryDialog = new TextHistoryDialog(this, this.model, this.actionMap);
    this.mainPanel.add(textHistoryDialog, TextHistoryDialog.class.toString());
    
  }
  
  private void setIcon() {
    
    String iconPath = KreawiGlobals.APP_IMAGES_DIR_NAME + "/trainer_icon.gif";
    
    URL url = TrainerMain.class.getResource(iconPath);
    setIconImage(new ImageIcon(url).getImage());
  }
  
  /**
   * Sets size to FRAME_SIZE and enlarge to the half of the free space
   * Sets location to the center of any available window area
   */
  private void setSizeAndLocation() {
    
    final Rectangle availableArea = WindowToolkit.getAvailableWindowArea(getGraphicsConfiguration());
    final double width = Math.min(availableArea.getWidth(), 
        FRAME_SIZE.getWidth() + (availableArea.getWidth() - FRAME_SIZE.getWidth()) * 0.5);
    final double height = Math.min(availableArea.getHeight(), 
        FRAME_SIZE.getHeight() + (availableArea.getHeight() - FRAME_SIZE.getHeight()) * 0.5);
    final double x = availableArea.getX() + (availableArea.getWidth() - width) / 2;
    final double y = availableArea.getY() + (availableArea.getHeight() - height) / 2;
    
    setSize(new Dimension((int) width, (int) height));
    setLocation((int) x, (int) y);
    getContentPane().setMaximumSize(FRAME_SIZE);
    mainPanel.setMaximumSize(FRAME_SIZE);
    
    if (this.model.getSettings().getBooleanValue(TrainerSettingsConstants.KEY_START_MAXIMIZED)) {
      setExtendedState(JFrame.MAXIMIZED_BOTH);
    }
    
  }
  
  private void switchDialog(String name) {
    this.dialogCards.show(this.mainPanel, name);
  }
  
  private JMenuBar buildMenuBar() {
    JMenuBar menuBar = new JMenuBar();
    menuBar.putClientProperty(PlasticLookAndFeel.IS_3D_KEY, true);
    menuBar.putClientProperty(Options.HEADER_STYLE_KEY, HeaderStyle.SINGLE);
    
    menuBar.add(createFileMenu());
    menuBar.add(createViewMenu());
    menuBar.add(createInfoMenu());
    
    return menuBar;
    
  }
  
  private JMenu createFileMenu() {
    JMenu fileMenu = new JMenu(TrainerTextResources.getString("MENU_TITLE_FILE"));
    
    // if (OSDetector.getOs().isMac()) {
    // /* mac os */
    // Application.getInstance().getPreferencesJMenuItem()
    // .addActionListener(this.actionMap.get(Actions.PREFERENCES));
    // } else {
    /* non-mac os */
    fileMenu.add(this.actionMap.get(Actions.PREFERENCES));
    fileMenu.addSeparator();
    // }
    
    fileMenu.add(this.actionMap.get(Actions.RESET_INTENSIVE_CARE));
    fileMenu.add(this.actionMap.get(Actions.RESET_NOTES));
    fileMenu.add(this.actionMap.get(Actions.RESET_QUESTION_ATTRIBUTES));
    
    fileMenu.addSeparator();
    fileMenu.add(this.actionMap.get(Actions.BACKUP_USER_DATA));
    fileMenu.add(this.actionMap.get(Actions.RESTORE_USER_DATA));
    
    if (KreawiGlobals.isDevMode()) {
      fileMenu.addSeparator();
      fileMenu.add(this.actionMap.get(Actions.EDIT_QUESTION));
    }
    
    if (OSDetector.getOs().isMac()) {
      /* mac os */
      fileMenu.setText(TrainerTextResources.getString("MENU_TITLE_FILE_MAC"));
      Application.getInstance().getQuitJMenuItem().addActionListener(this.actionMap.get(Actions.EXIT));
    } else {
      /* non-mac os */
      fileMenu.addSeparator();
      fileMenu.add(this.actionMap.get(Actions.EXIT));
    }
    
    return fileMenu;
  }
  
  private JMenu createInfoMenu() {
    JMenu menu = new JMenu(TrainerTextResources.getString("MENU_TITLE_INFO"));
    
    menu.add(this.actionMap.get(Actions.HELP_CONTENTS));
    menu.add(this.actionMap.get(Actions.TUTORIAL));
    menu.add(this.actionMap.get(Actions.TECHNICAL_SUPPORT));
    
    if (KreawiGlobals.isDevMode() || KreawiGlobals.isBetaMode()) {
      menu.add(this.actionMap.get(Actions.OPEN_LOG_DIR));
    }
    
    menu.addSeparator();
    
    JCheckBoxMenuItem checkOnStartupItem = new JCheckBoxMenuItem(
        this.actionMap.get(Actions.CHECK_FOR_UPDATE_ON_STARTUP));
    checkOnStartupItem.setSelected(checkOnStartupItem.getAction().getValue(ACTION_SELECTED_KEY).equals(Boolean.TRUE));
    
    menu.add(checkOnStartupItem);
    menu.add(this.actionMap.get(Actions.CHECK_FOR_UPDATES));
    
    // if (OSDetector.getOs().isMac()) {
    // /* mac os */
    // Application.getInstance().getAboutJMenuItem().addActionListener(
    // this.actionMap.get(Actions.ABOUT));
    // } else {
    /* non-mac os */
    
    menu.addSeparator();
    menu.add(this.actionMap.get(Actions.DELETE_LICENSE));
    
    menu.addSeparator();
    menu.add(this.actionMap.get(Actions.DELETE_ONLINE_GAME_PROFILE));

    menu.addSeparator();
    menu.add(this.actionMap.get(Actions.ABOUT));
    // }
    
    return menu;
  }
  
  private JMenu createViewMenu() {
    JMenu menu = new JMenu(TrainerTextResources.getString("MENU_TITLE_VIEW"));
    
    JCheckBoxMenuItem justifyItem = new JCheckBoxMenuItem(this.actionMap.get(Actions.JUSTIFY_TEXT));
    justifyItem.setSelected(justifyItem.getAction().getValue(ACTION_SELECTED_KEY).equals(Boolean.TRUE));
    
    menu.add(justifyItem);
    
    menu.addSeparator();
    
    JCheckBoxMenuItem showOnlineGameItem = new JCheckBoxMenuItem(this.actionMap.get(Actions.SHOW_ONLINE_GAME));
    showOnlineGameItem.setSelected(showOnlineGameItem.getAction().getValue(ACTION_SELECTED_KEY).equals(Boolean.TRUE));
    
    menu.add(showOnlineGameItem);
    
    return menu;
  }
  
  private void registerHelpItems() {
    
    HelpRegistry helpRegistry = this.model.getHelpRegistry();
    
    helpRegistry.registerHelpItem(TrainerSettingsConstants.KEY_SHOW_UNANSWERED_FILTER_EXPLANATION_MESSAGE,
        "SELECTION_LABEL_UNANSWERED_FILTER_EXPLANATION_TITLE", "SELECTION_LABEL_UNANSWERED_FILTER_EXPLANATION_MESSAGE");
    
    helpRegistry.registerHelpItem(TrainerSettingsConstants.KEY_SHOW_ICU_FILTER_EXPLANATION_MESSAGE,
        "SELECTION_LABEL_ICU_FILTER_EXPLANATION_TITLE", "SELECTION_LABEL_ICU_FILTER_EXPLANATION_MESSAGE");
    
    helpRegistry.registerHelpItem(TrainerSettingsConstants.KEY_SHOW_TEXT_SEARCH_EXPLANATION_MESSAGE,
        "SELECTION_LABEL_SEARCH_HELP_TITLE", "SELECTION_LABEL_SEARCH_HELP_MESSAGE");
    
    helpRegistry.registerHelpItem(TrainerSettingsConstants.KEY_SHOW_EXAM_SELECTION_EXPLANATION_MESSAGE,
        "SELECTION_LABEL_EXAM_SELECTION_EXPLANATION_TITLE", "SELECTION_LABEL_EXAM_SELECTION_EXPLANATION_MESSAGE");
    
    String[] sub = { TrainerTextResources.getString("HISTORY_BUTTON_LOAD"),
        TrainerTextResources.getString("HISTORY_BUTTON_DELETE"), TrainerTextResources.getString("HISTORY_LABEL_FILTER") };
    helpRegistry.registerHelpItem(TrainerSettingsConstants.KEY_SHOW_HISTORY_EXPLANATION_MESSAGE,
        "HISTORY_EXPLANATION_TITLE", "HISTORY_EXPLANATION_MESSAGE", sub);
    
    helpRegistry.registerHelpItem(TrainerSettingsConstants.KEY_SHOW_NOTES_EXPLANATION_MESSAGE,
        "QUIZ_SHOW_NOTES_EXPLANATION_TITLE", "QUIZ_SHOW_NOTES_EXPLANATION_MESSAGE");
    
    helpRegistry.registerHelpItem(TrainerSettingsConstants.KEY_SHOW_ICU_EXPLANATION_MESSAGE,
        "QUIZ_CHECKBOX_INTENSIVE_CARE_EXPLANATION_TITLE", "QUIZ_CHECKBOX_INTENSIVE_CARE_EXPLANATION_MESSAGE");
    
    String[] subs = { TrainerTextResources.getString("QUIZ_BUTTON_EVALUATION") };
    helpRegistry.registerHelpItem(TrainerSettingsConstants.KEY_SHOW_EXAM_MODE_EXPLANATION_MESSAGE,
        "SELECTION_LABEL_EXAM_MODE_EXPLANATION_TITLE", "SELECTION_LABEL_EXAM_MODE_EXPLANATION_MESSAGE", subs);
    
    final int defaultLevel = this.model.getSettings().getIntValue(TrainerSettingsConstants.KEY_DEFAULT_CARD_BOX_LEVEL) + 1;
    final int maxLevel = this.model.getSettings().getIntValue(TrainerSettingsConstants.KEY_MAX_CARD_BOX_LEVEL) + 1;
    
    final String[] substitutesCardBoxExplanation = { "" + defaultLevel, "" + (defaultLevel + 1), "" + maxLevel };
    
    helpRegistry.registerHelpItem(TrainerSettingsConstants.KEY_SHOW_CARD_BOX_FILTER_EXPLANATION_MESSAGE,
        "SELECTION_LABEL_CARD_BOX_FILTER_EXPLANATION_TITLE", "SELECTION_LABEL_CARD_BOX_FILTER_EXPLANATION_MESSAGE",
        substitutesCardBoxExplanation);
    
    String modifierKey = Util.getModifierKeyName();
    String[] subst = { modifierKey, modifierKey, modifierKey };
    
    helpRegistry.registerHelpItem(TrainerSettingsConstants.KEY_SHOW_CATEGORY_SELECTION_MC_EXPLANATION_MESSAGE,
        "SELECTION_LABEL_CATEGORY_SELECTION_MC_EXPLANATION_TITLE",
        "SELECTION_LABEL_CATEGORY_SELECTION_MC_EXPLANATION_MESSAGE", subst);
    
    helpRegistry.registerHelpItem(TrainerSettingsConstants.KEY_SHOW_CATEGORY_SELECTION_TEXT_EXPLANATION_MESSAGE,
        "SELECTION_LABEL_CATEGORY_SELECTION_TEXT_EXPLANATION_TITLE",
        "SELECTION_LABEL_CATEGORY_SELECTION_TEXT_EXPLANATION_MESSAGE", subst);
    
    helpRegistry.registerHelpItem(TrainerSettingsConstants.KEY_SHOW_ICU_SELECTION_EXPLANATION_MESSAGE,
        "SELECTION_LABEL_ICU_SELECTION_EXPLANATION_TITLE", "SELECTION_LABEL_ICU_SELECTION_EXPLANATION_MESSAGE", subst);
    
    helpRegistry.registerHelpItem(TrainerSettingsConstants.KEY_SHOW_ONLINE_GAME_SIGNUP_EXPLANATION_MESSAGE,
        "ONLINE_GAME_SIGNUP_EXPLANATION_TITLE", "ONLINE_GAME_SIGNUP_EXPLANATION_MESSAGE");
  }
  
  private void initActions() {
    
    this.actionMap = new HashMap<Actions, Action>();
    
    AbstractAction action;
    
    /* Training for written exams, selection by category */
    action = new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        showMCCategorySelection();
      }
    };
    action.putValue(Action.NAME, TrainerTextResources.getString("SELECTION_METHOD_CATEGORIES"));
    action.putValue(Action.SHORT_DESCRIPTION, TrainerTextResources.getString("SELECTION_METHOD_CATEGORIES_TOOL_TIP"));
    
    this.actionMap.put(Actions.WRITTEN_BY_CATEGORY, action);
    
    action.setEnabled(this.model.getMcQuestionsByCategories().getCategories().size() > 0);
    
    /* Training for written exams, selection by past exam */
    action = new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        showMCExamSelection();
      }
    };
    action.putValue(Action.NAME, TrainerTextResources.getString("SELECTION_METHOD_EXAMS"));
    action
        .putValue(Action.SHORT_DESCRIPTION, TrainerTextResources.getString("SELECTION_METHOD_EXAMS_WRITTEN_TOOL_TIP"));
    this.actionMap.put(Actions.WRITTEN_BY_EXAM, action);
    
    action.setEnabled(this.model.getMcQuestionsByExams().getExams().size() > 0);
    
    /* Training for oral exams, selection by category */
    action = new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        showTextCategorySelection();
      }
    };
    action.putValue(Action.NAME, TrainerTextResources.getString("SELECTION_METHOD_CATEGORIES"));
    action.putValue(Action.SHORT_DESCRIPTION, TrainerTextResources.getString("SELECTION_METHOD_CATEGORIES_TOOL_TIP"));
    this.actionMap.put(Actions.ORAL_BY_CATEGORY, action);
    
    action.setEnabled(this.model.getTextQuestionsByCategories().getCategories().size() > 0);
    
    /* Training for oral exams, selection by past exam */
    action = new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        showTextExamSelection();
      }
    };
    action.putValue(Action.NAME, TrainerTextResources.getString("SELECTION_METHOD_EXAMS"));
    action.putValue(Action.SHORT_DESCRIPTION, TrainerTextResources.getString("SELECTION_METHOD_EXAMS_ORAL_TOOL_TIP"));
    this.actionMap.put(Actions.ORAL_BY_EXAM, action);
    
    action.setEnabled(this.model.getTextQuestionsByExams().getExams().size() > 0);
    
    /* Back to main dialog */
    action = new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        backToMainDialogPerformed(Boolean.TRUE.equals(getValue(ACTION_PROPOSE_EVALUATION_KEY)));
        putValue(ACTION_PROPOSE_EVALUATION_KEY, Boolean.FALSE);
      }
    };
    action.putValue(Action.NAME, TrainerTextResources.getString("SELECTION_BUTTON_BACK"));
    action.putValue(ACTION_PROPOSE_EVALUATION_KEY, Boolean.FALSE);
    this.actionMap.put(Actions.BACK_TO_MAIN_DIALOG, action);
    
    /* To evaluation dialog */
    action = new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        toEvaluationDialogPerformed();
        actionMap.get(Actions.BACK_TO_MAIN_DIALOG).putValue(TrainerFrame.ACTION_PROPOSE_EVALUATION_KEY, Boolean.FALSE);
      }
    };
    action.putValue(Action.NAME, TrainerTextResources.getString("QUIZ_BUTTON_EVALUATION"));
    this.actionMap.put(Actions.TO_EVALUATION, action);
    
    /* back to quiz */
    action = new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        backToQuizPerformed();
      }
    };
    action.putValue(Action.NAME, TrainerTextResources.getString("EVALUATION_BUTTON_BACK_TO_QUIZ"));
    this.actionMap.put(Actions.BACK_TO_QUIZ, action);
    
    /* repeat incorrect answered questions */
    action = new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        repeatIncorrectPerformed();
      }
    };
    action.putValue(Action.NAME, TrainerTextResources.getString("EVALUATION_BUTTON_REPEAT_INCORRECT"));
    this.actionMap.put(Actions.REPEAT_INCORRECT, action);
    
    action = new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        final int cardBoxLevel = ((Integer) getValue(ACTION_CARD_BOX_LEVEL_KEY)).intValue();
        startCardBoxQuiz(cardBoxLevel);
      }
    };
    this.actionMap.put(Actions.START_CARD_BOX_QUIZ, action);
    
    /* check for updates dialog */
    action = new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        checkForUpdatesNow();
      }
    };
    action.putValue(Action.NAME, TrainerTextResources.getString("MENU_ITEM_CHECK_FOR_LATEST_VERSION"));
    this.actionMap.put(Actions.CHECK_FOR_UPDATES, action);
    
    /* technical support dialog */
    action = new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        technicalSupportNeededPerformed();
      }
    };
    action.putValue(Action.NAME, TrainerTextResources.getString("MENU_ITEM_TECHNICAL_SUPPORT"));
    this.actionMap.put(Actions.TECHNICAL_SUPPORT, action);
    
    /* open log directory */
    action = new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        try {
          KreawiGlobals.openLogDir();
        } catch (IOException e1) {
          showMessageDialog("Konnte Logging-Verzeichnis nicht öffnen!", JOptionPane.ERROR_MESSAGE);
        }
      }
    };
    action.putValue(Action.NAME, TrainerTextResources.getString("MENU_ITEM_OPEN_LOG_DIR"));
    this.actionMap.put(Actions.OPEN_LOG_DIR, action);
    
    /* To about dialog */
    action = new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        toAboutDialogPerformed();
      }
    };
    action.putValue(Action.NAME, TrainerTextResources.getString("MENU_ITEM_ABOUT"));
    this.actionMap.put(Actions.ABOUT, action);
    
    /* exit */
    action = new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        exit();
      }
      
    };
    action.putValue(Action.NAME, TrainerTextResources.getString("MENU_ITEM_EXIT"));
    this.actionMap.put(Actions.EXIT, action);
    
    /* justify text action */
    action = new AbstractAction(TrainerTextResources.getString("MENU_ITEM_JUSTIFY_TEXT")) {
      public void actionPerformed(ActionEvent e) {
        if (getValue(ACTION_SELECTED_KEY) == null || getValue(ACTION_SELECTED_KEY).equals(Boolean.FALSE)) {
          putValue(ACTION_SELECTED_KEY, Boolean.TRUE);
          model.getSettings().setBooleanValue(TrainerSettingsConstants.KEY_JUSTIFY_TEXT, true);
        } else if (getValue(ACTION_SELECTED_KEY).equals(Boolean.TRUE)) {
          putValue(ACTION_SELECTED_KEY, Boolean.FALSE);
          model.getSettings().setBooleanValue(TrainerSettingsConstants.KEY_JUSTIFY_TEXT, false);
        }
      }
    };
    action.putValue(ACTION_SELECTED_KEY,
        this.model.getSettings().getBooleanValue(TrainerSettingsConstants.KEY_JUSTIFY_TEXT));
    /*
     * workaround for older JRE versions, which do not render justifed html
     * correctly
     */
    String javaVersion = System.getProperty("java.version");
    try {
      if (javaVersion.startsWith("1.5.0_") && Integer.parseInt(javaVersion.substring(6)) < 11) {
        action.putValue(ACTION_SELECTED_KEY, false);
        action.setEnabled(false);
      }
    } catch (NumberFormatException nfe) {
      /* workaround failed */
      logger.info("Workaround for older JRE versions, which do not render justifed html correctly, has failed!");
    }
    this.actionMap.put(Actions.JUSTIFY_TEXT, action);
    
    
    /* show online game */
    action = new AbstractAction(TrainerTextResources.getString("MENU_ITEM_SHOW_ONLINE_GAME")) {
      public void actionPerformed(ActionEvent e) {
        if (getValue(ACTION_SELECTED_KEY) == null || getValue(ACTION_SELECTED_KEY).equals(Boolean.FALSE)) {
          putValue(ACTION_SELECTED_KEY, Boolean.TRUE);
          model.getSettings().setBooleanValue(TrainerSettingsConstants.KEY_SHOW_ONLINE_GAME, true);
        } else if (getValue(ACTION_SELECTED_KEY).equals(Boolean.TRUE)) {
          putValue(ACTION_SELECTED_KEY, Boolean.FALSE);
          model.getSettings().setBooleanValue(TrainerSettingsConstants.KEY_SHOW_ONLINE_GAME, false);
        }
        Platform.runLater(() -> {
          TrainerFrame.this.mainDialog.initRootPaneContent();
        });
      }
    };
    action.putValue(ACTION_SELECTED_KEY,
        this.model.getSettings().getBooleanValue(TrainerSettingsConstants.KEY_SHOW_ONLINE_GAME));
    this.actionMap.put(Actions.SHOW_ONLINE_GAME, action);
    
    /* Delete Online Game Profile */
    action = new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        deleteOnlineGameProfile();
      }
    };
    action.putValue(Action.NAME, TrainerTextResources.getString("MENU_ITEM_DELETE_ONLINE_GAME_PROFILE"));
    this.actionMap.put(Actions.DELETE_ONLINE_GAME_PROFILE, action);
    
    /* check for updates on startup action */
    action = new AbstractAction(TrainerTextResources.getString("MENU_ITEM_CHECK_FOR_UPDATES_ON_STARTUP")) {
      public void actionPerformed(ActionEvent e) {
        if (getValue(ACTION_SELECTED_KEY) == null || getValue(ACTION_SELECTED_KEY).equals(Boolean.FALSE)) {
          putValue(ACTION_SELECTED_KEY, Boolean.TRUE);
          model.getSettings().setBooleanValue(TrainerSettingsConstants.KEY_CHECK_FOR_UPDATE_ON_STARTUP, true);
        } else if (getValue(ACTION_SELECTED_KEY).equals(Boolean.TRUE)) {
          putValue(ACTION_SELECTED_KEY, Boolean.FALSE);
          model.getSettings().setBooleanValue(TrainerSettingsConstants.KEY_CHECK_FOR_UPDATE_ON_STARTUP, false);
        }
      }
    };
    action.putValue(ACTION_SELECTED_KEY,
        this.model.getSettings().getBooleanValue(TrainerSettingsConstants.KEY_CHECK_FOR_UPDATE_ON_STARTUP));
    this.actionMap.put(Actions.CHECK_FOR_UPDATE_ON_STARTUP, action);
    
    /* put all incorrect questions into Intensive Care Unit */
    action = new AbstractAction(TrainerTextResources.getString("EVALUATION_BUTTON_INCORRECT_TO_ICU")) {
      public void actionPerformed(ActionEvent e) {
        incorrectToIcuPerformed();
      }
    };
    this.actionMap.put(Actions.INCORRECT_QUESTIONS_INTO_ICU, action);
    
    /* Start MC Intensive Care Unit */
    action = new AbstractAction(TrainerTextResources.getString("SELECTION_METHOD_INTENSIVE_CARE")) {
      public void actionPerformed(ActionEvent e) {
        showMCICUSelection();
      }
    };
    action.putValue(Action.SHORT_DESCRIPTION,
        TrainerTextResources.getString("SELECTION_METHOD_INTENSIVE_CARE_TOOL_TIP"));
    this.actionMap.put(Actions.MC_INTENSIVE_CARE, action);
    
    /* Start text Intensive Care Unit */
    action = new AbstractAction(TrainerTextResources.getString("SELECTION_METHOD_INTENSIVE_CARE")) {
      public void actionPerformed(ActionEvent e) {
        showTextICUSelection();
      }
    };
    action.putValue(Action.SHORT_DESCRIPTION,
        TrainerTextResources.getString("SELECTION_METHOD_INTENSIVE_CARE_TOOL_TIP"));
    this.actionMap.put(Actions.TEXT_INTENSIVE_CARE, action);
    
    /* Show MC history */
    action = new AbstractAction(TrainerTextResources.getString("SELECTION_METHOD_HISTORY")) {
      public void actionPerformed(ActionEvent e) {
        showMcHistory();
      }
    };
    action.putValue(Action.SHORT_DESCRIPTION,TrainerTextResources.getString("SELECTION_METHOD_HISTORY_TOOL_TIP"));
    this.actionMap.put(Actions.MC_HISTORY, action);
    
    /* Show text history */
    action = new AbstractAction(TrainerTextResources.getString("SELECTION_METHOD_HISTORY")) {
      public void actionPerformed(ActionEvent e) {
        showTextHistory();
      }
    };
    action.putValue(Action.SHORT_DESCRIPTION,TrainerTextResources.getString("SELECTION_METHOD_HISTORY_TOOL_TIP"));
    this.actionMap.put(Actions.TEXT_HISTORY, action);
    
    /* Start random quiz with MC questions */
    action = new AbstractAction(TrainerTextResources.getString("SELECTION_METHOD_RANDOM_QUIZ")) {
      public void actionPerformed(ActionEvent e) {
        startMcRandomQuiz();
      }
    };
    action.putValue(Action.SHORT_DESCRIPTION, TrainerTextResources.getString("SELECTION_METHOD_RANDOM_QUIZ_TOOL_TIP"));
    this.actionMap.put(Actions.MC_RANDOM_QUIZ, action);
    
    /* Start random quiz with text questions */
    action = new AbstractAction(TrainerTextResources.getString("SELECTION_METHOD_RANDOM_QUIZ")) {
      public void actionPerformed(ActionEvent e) {
        startTextRandomQuiz();
      }
    };
    action.putValue(Action.SHORT_DESCRIPTION, TrainerTextResources.getString("SELECTION_METHOD_RANDOM_QUIZ_TOOL_TIP"));
    this.actionMap.put(Actions.TEXT_RANDOM_QUIZ, action);

    /* Remove all questions from Intensive Care Unit */
    action = new AbstractAction(TrainerTextResources.getString("MENU_ITEM_RESET_ICU")) {
      public void actionPerformed(ActionEvent e) {
        resetIcu();
      }
    };
    this.actionMap.put(Actions.RESET_INTENSIVE_CARE, action);
    
    /* Erase all personal notes */
    action = new AbstractAction(TrainerTextResources.getString("MENU_ITEM_RESET_NOTES")) {
      public void actionPerformed(ActionEvent e) {
        resetNotes();
      }
    };
    this.actionMap.put(Actions.RESET_NOTES, action);
    
    /* Reset flashcard box and answered questions */
    action = new AbstractAction(TrainerTextResources.getString("MENU_ITEM_RESET_QUESTION_ATTRIBUTES")) {
      public void actionPerformed(ActionEvent e) {
        resetCardBox();
      }
    };
    this.actionMap.put(Actions.RESET_QUESTION_ATTRIBUTES, action);
    
    /* Backup all user data */
    action = new AbstractAction(TrainerTextResources.getString("MENU_ITEM_BACKUP_USER_DATA")) {
      public void actionPerformed(ActionEvent e) {
        showBackupUserDataDialog();
      }
    };
    this.actionMap.put(Actions.BACKUP_USER_DATA, action);
    
    /* Restore previously backed up user data */
    action = new AbstractAction(TrainerTextResources.getString("MENU_ITEM_RESTORE_USER_DATA")) {
      public void actionPerformed(ActionEvent e) {
        showRestoreUserDataDialog();
      }
    };
    this.actionMap.put(Actions.RESTORE_USER_DATA, action);
    
    /* Edit question (only for dev mode!) */
    action = new AbstractAction(TrainerTextResources.getString("MENU_ITEM_EDIT_QUESTION")) {
      public void actionPerformed(ActionEvent e) {
        editQuestionPerformed();
      }
    };
    action.putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_E, Event.CTRL_MASK));
    this.actionMap.put(Actions.EDIT_QUESTION, action);
    
    /* Show help dialog */
    action = new AbstractAction(TrainerTextResources.getString("MENU_ITEM_HELP_CONTENTS")) {
      public void actionPerformed(ActionEvent e) {
        showHelpDialog();
      }
    };
    if (OSDetector.getOs().isMac()) {
      /* mac os */
      int shortcut = Toolkit.getDefaultToolkit().getMenuShortcutKeyMask();
      action.putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke('?', shortcut));
    } else {
      /* non-mac os */
      action.putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_F1, 0));
    }
    this.actionMap.put(Actions.HELP_CONTENTS, action);
    
    /* Show preferences dialog */
    action = new AbstractAction(TrainerTextResources.getString("MENU_ITEM_PREFERENCES")) {
      public void actionPerformed(ActionEvent e) {
        showPreferencesDialog();
      }
    };
    this.actionMap.put(Actions.PREFERENCES, action);
    
    action = new AbstractAction(TrainerTextResources.getString("MENU_ITEM_DELETE_LICENSE")) {
      public void actionPerformed(ActionEvent e) {
        confirmDeleteLicense();
      }
    };
    this.actionMap.put(Actions.DELETE_LICENSE, action);
    
    final TrainerFrame trainerFrame = this;
    action = new AbstractAction(TrainerTextResources.getString("MENU_ITEM_TUTORIAL")) {
      public void actionPerformed(ActionEvent e) {
        TutorialHandler.showRandomTip(new Config(model.getSettings().getSettingProperties()), trainerFrame);
      }
    };
    this.actionMap.put(Actions.TUTORIAL, action);
  }
  
  private void showMCCategorySelection() {
    switchDialog(MCCategorySelectionDialog.class.toString());
  }
  
  private void showMCExamSelection() {
    switchDialog(ExamSelection.class.toString());
  }
  
  private void showMCICUSelection() {
    switchDialog(MCICUSelectionDialog.class.toString());
  }
  
  private void showTextCategorySelection() {
    switchDialog(TextCategorySelectionDialog.class.toString());
  }
  
  private void showTextExamSelection() {
    switchDialog(ExamSelection.class.toString() + Question.Type.TEXT);
  }
  
  private void showTextICUSelection() {
    switchDialog(TextICUSelectionDialog.class.toString());
  }
  
  protected void startMcRandomQuiz() {
    final QuestionPool mcQuestions = model.getAllMcQuestions();
    final String title = TrainerTextResources.getString("RANDOM_QUIZ_CONFIRM_TITLE");
    final String message = TrainerTextResources.getString("RANDOM_QUIZ_ASK_EXAM_MODE");
    String[] options = { TrainerTextResources.getString("RANDOM_QUIZ_EXAM_MODE_YES"),
        TrainerTextResources.getString("RANDOM_QUIZ_EXAM_MODE_NO"),
        TrainerTextResources.getString("RANDOM_QUIZ_EXAM_MODE_CANCEL") };
    final int result = JOptionPane.showOptionDialog(this, message, title, JOptionPane.YES_NO_CANCEL_OPTION,
        JOptionPane.QUESTION_MESSAGE, null, options, options[0]);
    if (result != JOptionPane.CANCEL_OPTION) {
      final boolean examMode = (result == JOptionPane.YES_OPTION);
      startRandomQuiz(mcQuestions, examMode);
    } 
  }
  
  protected void startTextRandomQuiz() {
    final QuestionPool textQuestions = model.getAllTextQuestions();
    final boolean examMode = false; // macht keinen Sinn für Text-Fragen
    final String title = TrainerTextResources.getString("RANDOM_QUIZ_CONFIRM_TITLE");
    final String message = TrainerTextResources.getString("RANDOM_QUIZ_CONFIRM_TEXT_QUIZ");
    final int result = JOptionPane.showConfirmDialog(this, message, title, JOptionPane.OK_CANCEL_OPTION,
        JOptionPane.QUESTION_MESSAGE);
    if (result == JOptionPane.OK_OPTION) {
      startRandomQuiz(textQuestions, examMode);
    }
  }

  private void startRandomQuiz(final QuestionPool allQuestions, final boolean examMode) {
    final ArrayList<TitledQuestionGroup> groups = new ArrayList<TitledQuestionGroup>(); 
    groups.addAll(allQuestions.getCategories());
    groups.addAll(allQuestions.getExams());
    
    final SwingWorker<KreawiQuiz, Void> worker = new SwingWorker<KreawiQuiz, Void>() {
      public KreawiQuiz doInBackground() {
        final QuestionPool selectedQuestions = allQuestions.getRandomQuestionPoolForGroups(groups, 60);
        final KreawiQuiz quiz = new KreawiQuiz(selectedQuestions, examMode);
        quiz.sortQuestionsById();
        return quiz;
      }
      public void done() {
        try {
          model.setSelectedQuestions(get());
        } catch (InterruptedException e) {
          logger.error("Erstellen der Zufallsprüfung unterbrochen!", e);
        } catch (ExecutionException e) {
          logger.error("Fehler beim Erstellen der Zufallsprüfung!", e);
          showFatalErrorDialogWithoutQuit();
        } catch (CancellationException e) {
          logger.info("Erstellen der Zufallsprüfung wurde vom Benutzer abgebrochen!");
        }
      }
    };
    final String waitMessage = TrainerTextResources.getString("RANDOM_QUIZ_LOADING");
    showPleaseWaitDialog(waitMessage, worker);
  }

  private void backToMainDialogPerformed(boolean proposeEvaluation) {
    
    if (proposeEvaluation && evaluationDesired()) {
      toEvaluationDialogPerformed();
    } else if (reallyDiscard(TrainerTextResources.getString("QUIZ_CONFIRM_BACK_TO_MAIN_MESSAGE"),
        TrainerTextResources.getString("QUIZ_CONFIRM_BACK_TO_MAIN_TITLE"), true)) {
      possiblyWriteQuizToDisk();
      discard();
      switchDialog(MainDialogFx.class.toString());
    }
    
  }
  
  private void startCardBoxQuiz(final int cardBoxLevel) {
    final QuestionPool questions = model.getQuestionsForCardBoxLevel(cardBoxLevel);
    if (questions.getQuestionCount() == 0) {
      final String message = TrainerTextResources.getString("CARD_BOX_COMPARTMENT_EMPTY");
      showMessageDialog(message, JOptionPane.INFORMATION_MESSAGE);
      return;
    }
    final int maxCardBoxLevel = this.model.getSettings().getIntValue(TrainerSettingsConstants.KEY_MAX_CARD_BOX_LEVEL);
    final String compartment;
    if (cardBoxLevel == maxCardBoxLevel) {
      compartment = TrainerTextResources.getString("CARD_BOX_COMPARTMENT_ARCHIVE");
    } else {
      compartment = "" + (cardBoxLevel + 1);  
    }
    
    final String message = TrainerTextResources.getStringWithSubstitution("CARD_BOX_REPEAT_COMPARTMENT", compartment);
    final int result = JOptionPane.showConfirmDialog(this, message, null, JOptionPane.YES_NO_OPTION);
    if (result != JOptionPane.YES_OPTION) {
      return;
    }
    final KreawiQuiz quiz = new KreawiQuiz(questions, false, cardBoxLevel);
    
    // Sortierung nach Einsortierung der Fragen
    Comparator<Question> comp = new Comparator<Question>() {
      public int compare(Question o1, Question o2) {
        Long timestamp1 = Long.valueOf(model
            .getCardBoxTimestamp(o1));
        Long timestamp2 = Long.valueOf(model
            .getCardBoxTimestamp(o2));

        return timestamp1.compareTo(timestamp2);
      }
    };

    quiz.sortQuestions(comp);
    
    possiblyWriteQuizToDisk();
    discard();
    model.setSelectedQuestions(quiz);
  }

  private void backToQuizPerformed() {
    
    if (this.mcDialog != null) {
      this.mainPanel.add(this.mcDialog, QuizDialog.class.toString());
      switchDialog(QuizDialog.class.toString());
    }
    
  }
  
  private void toEvaluationDialogPerformed() {
    
    if (this.evalDialog == null) {
      this.evalDialog = new EvaluationDialog(this, this.model, this.actionMap);
      
      this.mainPanel.add(this.evalDialog, EvaluationDialog.class.toString());
    }
    
    if (this.mcDialog != null) {
      this.mainPanel.remove(this.mcDialog);
    }
    switchDialog(EvaluationDialog.class.toString());
    
  }
  
  private void repeatIncorrectPerformed() {
    
    if (reallyDiscard(TrainerTextResources.getString("QUIZ_CONFIRM_REPEAT_INCORRECT_MESSAGE"),
        TrainerTextResources.getString("QUIZ_CONFIRM_REPEAT_INCORRECT_TITLE"), false)) {
      
      final QuestionPool incorrect = this.model.getAllQuestions().getEmptyQuestionPool();
      final KreawiQuiz oldQuiz = this.model.getSelectedQuestions();
      incorrect.addAllQuestions(oldQuiz.getIncorrectAnsweredQuestions());
      final boolean isExamMode = oldQuiz.isExamMode();
      final KreawiQuiz newQuiz;
      if (oldQuiz.isCardBoxSelection()) {
        final int cardBoxLevel = oldQuiz.cardBoxLevel();
        newQuiz = new KreawiQuiz(incorrect, isExamMode, cardBoxLevel);
      } else {
        newQuiz = new KreawiQuiz(incorrect, isExamMode);
      }
      possiblyWriteQuizToDisk();
      discard();
      this.model.setSelectedQuestions(newQuiz);
    }
    
  }
  
  private void incorrectToIcuPerformed() {
    
    QuestionPool incorrect = this.model.getAllQuestions().getEmptyQuestionPool();
    incorrect.addAllQuestions(this.model.getSelectedQuestions().getIncorrectAnsweredQuestions());
    
    this.model.addAllQuestionsToIcu(incorrect);
    
  }
  
  private void toAboutDialogPerformed() {
    
    File trainerLicense = new File("Lizenz.txt");
    File jGoodiesLooksLicense = new File("lib/JGoodies_Looks_LICENSE.txt");
    File jGoodiesFormsLicense = new File("lib/JGoodies_Forms_LICENSE.txt");
    File mrjAdapterLicense = new File("lib/MRJAdapter_license.txt");
    File jlGuiLicense = new File("lib/jlGui/LICENSE.txt");
    
    String trainerLicenseString = "";
    String jGoodiesLooksLicenseString = "";
    String jGoodiesFormsLicenseString = "";
    String mrjAdapterLicenseString = "";
    String jlGuiLicenseString = "";
    
    try {
      trainerLicenseString = "<a href=\"" + trainerLicense.toURI().toURL().toString() + "\">"
          + TrainerTextResources.getString("LICENSE") + "</a>";
      jGoodiesLooksLicenseString = "(<a href=\"" + jGoodiesLooksLicense.toURI().toURL().toString() + "\">"
          + TrainerTextResources.getString("LICENSE") + "</a>) "
          + LookUtils.class.getPackage().getImplementationVersion();
      jGoodiesFormsLicenseString = "(<a href=\"" + jGoodiesFormsLicense.toURI().toURL().toString() + "\">"
          + TrainerTextResources.getString("LICENSE") + "</a>) "
          + FormUtils.class.getPackage().getImplementationVersion();
      mrjAdapterLicenseString = "(<a href=\"" + mrjAdapterLicense.toURI().toURL().toString() + "\">"
          + TrainerTextResources.getString("LICENSE") + "</a>)";
      jlGuiLicenseString = "(<a href=\"" + jlGuiLicense.toURI().toURL().toString() + "\">"
          + TrainerTextResources.getString("LICENSE") + "</a>)";
    } catch (MalformedURLException e) {
      logger.error("Could not generate some infos!", e);
    }
    
    String browserLauncherVersionString = BrowserLauncher.class.getPackage().getImplementationVersion();
    if (browserLauncherVersionString == null) {
      browserLauncherVersionString = "";
    }
    
    final String suffix;
    if (KreawiGlobals.isDemo()) {
      suffix = TrainerTextResources.getString("APPLICATION_TITLE_SUFFIX_DEMO");
    } else if (KreawiGlobals.isBetaMode()) {
      suffix = TrainerTextResources.getString("APPLICATION_TITLE_SUFFIX_BETA");
    } else {
      suffix = "";
    }

    final String[] substitutes = { KreawiGlobals.CURRENT_YEAR,
        Util.getVersion(TrainerMain.class) == null ? "N/A" : Util.getVersion(TrainerMain.class).toString(),
        suffix,
        KreawiGlobals.CURRENT_YEAR, KreawiGlobals.CURRENT_YEAR, trainerLicenseString, jGoodiesLooksLicenseString,
        jGoodiesFormsLicenseString, browserLauncherVersionString, mrjAdapterLicenseString, jlGuiLicenseString };
    
    final String text = TrainerTextResources.getStringWithSubstitution("ABOUT_DIALOG_TEXT", substitutes);
    JTextPane textPane = TrainerHtmlUtil.createHtmlPane(text);
    textPane.setOpaque(true);
    
    JScrollPane scrollPane = Factory.createStrippedScrollPane(textPane);
    scrollPane.setPreferredSize(SMALL_DIALOG_SIZE);
    scrollPane.setBorder(BorderFactory.createEtchedBorder(EtchedBorder.LOWERED));
    
    JOptionPane.showMessageDialog(this, scrollPane, TrainerTextResources.getString("ABOUT_DIALOG_TITLE"),
        JOptionPane.PLAIN_MESSAGE, new ImageIcon(TrainerMain.class.getResource(MainDialogFx.LOGO_PATH)));
    
  }
  
  private void showHelpDialog() {
    
    String close = TrainerTextResources.getString("HELP_DIALOG_BACK");
    String[] options = { close };
    JOptionPane.showOptionDialog(this, new HelpDialog(this.model.getHelpRegistry()),
        TrainerTextResources.getString("MENU_ITEM_HELP_CONTENTS"), JOptionPane.YES_NO_OPTION,
        JOptionPane.PLAIN_MESSAGE, null, options, close);
    
  }
  
  private void showPreferencesDialog() {
    final Properties tempProperties = this.model.getSettings().getSettingProperties();
    ConfigComponentHandler configCompHandler = new ConfigComponentHandler();
    if (JOptionPane.showOptionDialog(this, new PreferencesDialog(this.model.getConfig(), new Config(tempProperties),
        configCompHandler), TrainerTextResources.getString("SETTINGS_DIALOG_TITLE"), JOptionPane.OK_CANCEL_OPTION,
        JOptionPane.PLAIN_MESSAGE, null, null, null) == JOptionPane.OK_OPTION) {
      configCompHandler.commit();
    }
  }
  
  private void checkForUpdatesNow() {
    if (KreawiGlobals.isDemo()) {
      final String demoTitle = TrainerTextResources.getString("NOT_IN_DEMO_TITLE");
      final String demoText = TrainerTextResources.getString("NOT_IN_DEMO_TEXT");
      showMessageDialog(demoTitle, demoText, JOptionPane.INFORMATION_MESSAGE);
      return;
    }
    getLatestVersionHandler().showLatestVersionDialog();
  }
  
  private void technicalSupportNeededPerformed() {
    
    if (JOptionPane.YES_OPTION == (JOptionPane.showConfirmDialog(this,
        TrainerTextResources.getString("TECHNICAL_SUPPORT_TEXT"),
        TrainerTextResources.getString("TECHNICAL_SUPPORT_TITLE"), JOptionPane.YES_NO_OPTION,
        JOptionPane.PLAIN_MESSAGE, new ImageIcon(TrainerMain.class.getResource(MainDialogFx.LOGO_PATH))))) {
      
      final String url = TrainerTextResources.getString("TECHNICAL_SUPPORT_URL");
      try {
        logger.info("Öffne URL: " + url);
        DesktopApi.browse(new URI(url));
      } catch (Exception ex) {
        logger.error("Desktop.browse hat nicht funktioniert! (url: " + url + ")", ex);
        try {
          (new BrowserLauncher()).openURLinBrowser(url);
        } catch (Exception ex2) {
          logger.error("Konnte URL nicht öffnen! (url: " + url + ")", ex2);
          final String message = TrainerTextResources.getString("TECHNICAL_SUPPORT_FAILED_TEXT");
          final String title = TrainerTextResources.getString("TECHNICAL_SUPPORT_FAILED_TITLE");
          showMessageDialog(title, message, JOptionPane.WARNING_MESSAGE);
        }
      }
    }
    
  }
  
  private void confirmDeleteLicense() {
    if (KreawiGlobals.isDemo()) {
      final String demoTitle = TrainerTextResources.getString("NOT_IN_DEMO_TITLE");
      final String demoText = TrainerTextResources.getString("NOT_IN_DEMO_TEXT");
      showMessageDialog(demoTitle, demoText, JOptionPane.INFORMATION_MESSAGE);
      return;
    }
    final String title = TrainerTextResources.getString("DELETE_LICENSE_TITLE");
    final String textConfirm = TrainerTextResources.getString("DELETE_LICENSE_TEXT");
    final Component parent = this;
    SwingUtilities.invokeLater(new Runnable() {
      public void run() {
        final String deleteLicenseYes = TrainerTextResources.getString("DELETE_LICENSE_YES");
        final String deleteLicenseNo = TrainerTextResources.getString("DELETE_LICENSE_NO");
        final String[] deleteLicenseOptions = new String[] { deleteLicenseYes, deleteLicenseNo };
        final int answer = JOptionPane.showOptionDialog(parent, textConfirm, title, JOptionPane.YES_NO_OPTION,
            JOptionPane.QUESTION_MESSAGE, null, deleteLicenseOptions, deleteLicenseYes);
        if (answer == JOptionPane.YES_OPTION) {
          
          final Runnable deleteLicense = new Runnable() {
            public void run() {
              KreawiLicenseHandler.getInstance().deleteLicense();
              final Runnable closeTrainer = new Runnable() {
                public void run() {
                  exit();
                }
              };
              final String textSuccess = TrainerTextResources.getString("DELETE_LICNESE_SUCCESS");
              showMessageDialog(title, textSuccess, JOptionPane.INFORMATION_MESSAGE, closeTrainer);
            }
          };

          if (RegistrationHandler.getInstance().deleteRegistration()) {
            deleteLicense.run();
          } else {
            final String textDelRegFailed = TrainerTextResources.getString("DELETE_LICENSE_ANYWAY_TEXT");
            final String deleteAnywayYes = TrainerTextResources.getString("DELETE_LICENSE_ANYWAY_YES");
            final String deleteAnywayNo = TrainerTextResources.getString("DELETE_LICENSE_ANYWAY_NO");
            final String[] deleteAnywayOptions = new String[] { deleteAnywayYes, deleteAnywayNo };
            final int deleteAnywayAnswer = JOptionPane.showOptionDialog(parent, textDelRegFailed, title, JOptionPane.YES_NO_OPTION,
                JOptionPane.WARNING_MESSAGE, null, deleteAnywayOptions, deleteAnywayYes);
            if (deleteAnywayAnswer == JOptionPane.YES_OPTION) {
              deleteLicense.run();
            }
          }
        }
      }
    });
  }
  
  protected void deleteOnlineGameProfile() {
    final String title = TrainerTextResources.getString("DELETE_ONLINE_GAME_PROFILE_CONFIRM_TITLE");
    final String textConfirm = TrainerTextResources.getString("DELETE_ONLINE_GAME_PROFILE_CONFIRM_MESSAGE");
    final Component parent = this;
    SwingUtilities.invokeLater(() -> {
      final String yes = TrainerTextResources.getString("DELETE_LICENSE_YES");
      final String no = TrainerTextResources.getString("DELETE_LICENSE_NO");
      final String[] options = new String[] { yes, no };
      final int answer = JOptionPane.showOptionDialog(parent, textConfirm, title, JOptionPane.YES_NO_OPTION,
          JOptionPane.QUESTION_MESSAGE, null, options, yes);
      if (answer == JOptionPane.YES_OPTION) {
        
        final Runnable onSuccess = () -> {
          if (model.getSettings().getBooleanValue(TrainerSettingsConstants.KEY_SHOW_ONLINE_GAME)) {
            model.getSettings().setBooleanValue(TrainerSettingsConstants.KEY_SHOW_ONLINE_GAME, false);
            Platform.runLater(() -> {
              TrainerFrame.this.mainDialog.initRootPaneContent();
            });
          }
          SwingUtilities.invokeLater(() -> {
            final String textSuccess = TrainerTextResources.getString("DELETE_ONLINE_GAME_PROFILE_SUCCESS");
            showMessageDialog(title, textSuccess, JOptionPane.INFORMATION_MESSAGE);
          });
        };
        
        final ServerFailureCallback onFailure = new ServerFailureCallback() {
          @Override
          public void run(String serverMessage) {
            SwingUtilities.invokeLater(() -> {
              final String textFailure = TrainerTextResources
                  .getStringWithSubstitution("DELETE_ONLINE_GAME_PROFILE_FAILURE", serverMessage);
              showMessageDialog(title, textFailure, JOptionPane.ERROR_MESSAGE);
            });
          }
        };
        
        this.model.getOnlineGame().deleteProfile(onSuccess, onFailure);
        
      }
    });
    
  }

  private void editQuestionPerformed() {
    
    // try {
    // ArrayList<Integer> wrap = new ArrayList<Integer>();
    // wrap.add(this.model.getSelectedQuestions().getCurrentQuestion()
    // .getId());
    //
    // new QuestionTool(wrap);
    // } catch (ClassNotFoundException e) {
    // displayErrorMessage(e);
    // } catch (SQLException e) {
    // displayErrorMessage(e);
    // } catch (IOException e) {
    // displayErrorMessage(e);
    // } catch (NumberFormatException e) {
    // displayErrorMessage(e);
    // } catch (NullPointerException e) {
    // displayErrorMessage(e);
    // }
    
  }
  
  // private void displayErrorMessage(Exception e) {
  // JOptionPane.showMessageDialog(this, "Exception in thread \""
  // + Thread.currentThread().getName() + "\"\n" + e,
  // "Exception occurred!", JOptionPane.ERROR_MESSAGE);
  // log("QuestionCheck could not be launched due to an exception!", e);
  // }
  
  public void exit() {
    if (m_pleaseWaitDialog != null && m_pleaseWaitDialog.isVisible()) {
      logger.info("Close-Window-Button pressed while closing! Canceling current action!");
      m_pleaseWaitDialog.cancel();
      return;
    }
    
    if (!reallyDiscard(TrainerTextResources.getString("QUIZ_CONFIRM_EXIT_MESSAGE"),
        TrainerTextResources.getString("QUIZ_CONFIRM_EXIT_TITLE"), true)) {
      return;
    }
    
    if (this.m_closing) {
      logger.warn("TrainerFrame.exit() called twice!!!");
      return;
    }
    
    this.m_closing = true;
    
    final boolean forcedShutdown = TrainerUtil.isShuttingDown();
    
    try {
      if (forcedShutdown) {
        logger.warn("Detected forced shutdown. Not saving current quiz!");
      } else {
        possiblyWriteQuizToDisk();
        if (this.getExtendedState() == JFrame.MAXIMIZED_BOTH) {
          this.model.getSettings().setBooleanValue(TrainerSettingsConstants.KEY_START_MAXIMIZED, true);
        } else {
          this.model.getSettings().setBooleanValue(TrainerSettingsConstants.KEY_START_MAXIMIZED, false);
        }
      }
      main.exit(forcedShutdown);
      
    } catch (Exception e) {
      logger.error("Error while closing! Forcing shutdown!", e);
      main.exit(true);
    }
    
    
  }
  
  public boolean evaluationDesired() {
    
    String[] options = { TrainerTextResources.getString("QUIZ_BUTTON_EVALUATION"),
        TrainerTextResources.getString("SELECTION_BUTTON_BACK") };
    
    return JOptionPane.showOptionDialog(this, TrainerTextResources.getString("QUIZ_COMPLETE_BACK_TO_MAIN_MESSAGE"),
        TrainerTextResources.getString("QUIZ_COMPLETE_BACK_TO_MAIN_TITLE"), JOptionPane.YES_NO_OPTION,
        JOptionPane.QUESTION_MESSAGE, null, options, options[0]) == JOptionPane.YES_OPTION;
    
  }
  
  private void resetIcu() {
    if (JOptionPane.showConfirmDialog(this, TrainerTextResources.getString("RESET_ICU_CONFIRM_MESSAGE"),
        TrainerTextResources.getString("RESET_ICU_CONFIRM_TITLE"), JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION) {
      this.model.resetIcu();
    }
  }
  
  private void resetNotes() {
    if (JOptionPane.showConfirmDialog(this, TrainerTextResources.getString("RESET_NOTES_CONFIRM_MESSAGE"),
        TrainerTextResources.getString("RESET_NOTES_CONFIRM_TITLE"), JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION) {
      this.model.resetNotes();
    }
  }
  
  private void resetCardBox() {
    if (JOptionPane.showConfirmDialog(this,
        TrainerTextResources.getString("RESET_QUESTION_ATTRIBUTES_CONFIRM_MESSAGE"),
        TrainerTextResources.getString("RESET_QUESTION_ATTRIBUTES_CONFIRM_TITLE"),
        JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION) {
      this.model.resetCardBox();
    }
  }
  
  private void showBackupUserDataDialog() {
    final String message = TrainerTextResources.getString("BACKUP_USER_DATA_MESSAGE");
    final String title = TrainerTextResources.getString("BACKUP_USER_DATA_TITLE");
    final String[] buttonTexts = new String[] { TrainerTextResources.getString("BACKUP_USER_DATA_CONFIRM"),
        TrainerTextResources.getString("BACKUP_USER_DATA_BACK") };
    final String backupTitle = JOptionPane.showInputDialog(this, message, title, JOptionPane.PLAIN_MESSAGE);
    if (backupTitle == null) {
      // canceled by user
      return;
    }
    
    final SwingWorker<Void, Void> worker = new SwingWorker<Void, Void>() {
      @Override
      protected Void doInBackground() throws Exception {
        model.backupUserData(backupTitle);
        return null;
      }
      public void done() {
        try {
          get();   
          final String successMessage = TrainerTextResources.getString("BACKUP_USER_DATA_SUCCESS");
          showMessageDialog(title, successMessage, JOptionPane.INFORMATION_MESSAGE);
        } catch (ExecutionException | InterruptedException ex) {
          // ex.getCause() would be a TrainerStorageException
          final String errorMessage = TrainerTextResources.getString("BACKUP_USER_DATA_ERROR");
          showMessageDialog(title, errorMessage, JOptionPane.ERROR_MESSAGE);
        } finally {
          CursorToolkit.stopWaitCursor(TrainerFrame.this.getRootPane());
        }
      }
    };
    CursorToolkit.startWaitCursor(TrainerFrame.this.getRootPane());
    worker.execute();
  }
  
  private void showRestoreUserDataDialog() {
    final String title = TrainerTextResources.getString("RESTORE_USER_DATA_TITLE");
    try {
      final TrainerBackup[] backups = model.readBackups();
      if (backups.length > 0) {
        final String message = TrainerTextResources.getString("RESTORE_USER_DATA_CHOOSE_BACKUP");
        final TrainerBackup selectedBackup = (TrainerBackup) JOptionPane.showInputDialog(this, message, title,
            JOptionPane.PLAIN_MESSAGE, null, backups, backups[0]);
        if (selectedBackup != null) {
          // canceled by user
          restoreBackup(selectedBackup);
        }
      } else {
        final String message = TrainerTextResources.getString("RESTORE_USER_DATA_NO_BACKUPS_FOUND");
        showMessageDialog(title, message, JOptionPane.WARNING_MESSAGE);
      }
    } catch (TrainerStorageException e) {
      final String message = TrainerTextResources.getString("RESTORE_USER_DATA_READ_BACKUPS_ERROR");
      showMessageDialog(title, message, JOptionPane.ERROR_MESSAGE);
    }
    
  }
  
  private void restoreBackup(TrainerBackup selectedBackup) {
    final String title = TrainerTextResources.getString("RESTORE_USER_DATA_TITLE");
    
    final SwingWorker<Void, Void> worker = new SwingWorker<Void, Void>() {
      @Override
      protected Void doInBackground() throws Exception {
        TrainerFrame.this.model.restoreBackup(selectedBackup);
        return null;
      }
      public void done() {
        try {
          get();   
          final String message = TrainerTextResources.getString("RESTORE_USER_DATA_SUCCESS");
          showMessageDialog(title, message, JOptionPane.INFORMATION_MESSAGE);
        } catch (ExecutionException | InterruptedException ex) {
          // ex.getCause() would be a TrainerStorageException
          final String message = TrainerTextResources.getString("RESTORE_USER_DATA_RESTORING_ERROR");
          showMessageDialog(title, message, JOptionPane.ERROR_MESSAGE);
        } finally {
          CursorToolkit.stopWaitCursor(TrainerFrame.this.getRootPane());
        }
      }
    };
    CursorToolkit.startWaitCursor(this.getRootPane());
    worker.execute();

  }
  
  public boolean reallyDiscard(String message, String title, boolean checkUnanswered) {
    if (this.model.getSelectedQuestions() != null && this.model.getSelectedQuestions().getAnsweredQuestionCount() > 0) {
      
      if (!checkUnanswered || this.model.getSelectedQuestions().hasUnanswered(false)) {
        
        return JOptionPane.showConfirmDialog(this, message, title, JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION;
        
      }
      
    }
    return true;
  }
  
  private void discard() {
    
    try {
      this.mainPanel.remove(this.mcDialog);
      this.mainPanel.remove(this.evalDialog);
    } catch (NullPointerException npe) {
      /* do nothing */
    }
    
    if (this.evalDialog != null) {
      this.evalDialog.closeDialog();
    }
    
    this.evalDialog = null;
    
    if (this.mcDialog != null) {
      this.mcDialog.closeDialog();
    }
    
    this.mcDialog = null;
    
    if (this.model.getSelectedQuestions() != null) {
      this.model.getSelectedQuestions().deleteObservers();
    }
    this.model.setSelectedQuestions(null);
    
  }
  
  private void possiblyWriteQuizToDisk() {
    KreawiQuiz quiz = this.model.getSelectedQuestions();
    
    if (quiz == null) {
      return;
    }
    
    try {
      this.model.writeQuizToDisk(quiz);
    } catch (TrainerStorageException e) {
      if (KreawiGlobals.isDevMode()) {
        showFatalErrorDialogWithoutQuit();
      }
    }
  }
  
  private void showMcHistory() {
    switchDialog(McHistoryDialog.class.toString());
  }
  
  private void showTextHistory() {
    switchDialog(TextHistoryDialog.class.toString());
  }
  
  public static enum Actions {
    WRITTEN_BY_CATEGORY, 
    WRITTEN_BY_EXAM, 
    ORAL_BY_CATEGORY, 
    ORAL_BY_EXAM, 
    BACK_TO_MAIN_DIALOG, 
    TO_EVALUATION, ABOUT, 
    EXIT, CHECK_FOR_UPDATES, 
    BACK_TO_QUIZ, 
    START_CARD_BOX_QUIZ, 
    NEW_QUESTION_SELECTED, 
    JUSTIFY_TEXT,
    SHOW_ONLINE_GAME,
    DELETE_ONLINE_GAME_PROFILE,
    CHECK_FOR_UPDATE_ON_STARTUP, 
    TECHNICAL_SUPPORT,
    OPEN_LOG_DIR,
    REPEAT_INCORRECT, 
    INCORRECT_QUESTIONS_INTO_ICU, 
    TEXT_INTENSIVE_CARE, MC_INTENSIVE_CARE, 
    RESET_INTENSIVE_CARE, 
    RESET_NOTES, 
    RESET_QUESTION_ATTRIBUTES, 
    EDIT_QUESTION, 
    HELP_CONTENTS, 
    PREFERENCES, 
    MC_HISTORY, 
    TEXT_HISTORY, 
    BACKUP_USER_DATA, 
    RESTORE_USER_DATA, 
    MC_RANDOM_QUIZ, 
    TEXT_RANDOM_QUIZ, 
    DELETE_LICENSE, 
    TUTORIAL;
  }
  
  private void startNewQuiz() {
    
    /* New question were selected, training is to be started. */
    
    this.mcDialog = new QuizDialog(this, this.model, this.actionMap);
    
    this.mcDialog.setSize(FRAME_SIZE);
    
    this.mainPanel.add(this.mcDialog, QuizDialog.class.toString());
    
    this.model.getSelectedQuestions().addObserver(new Observer() {
      public void update(Observable o, Object arg) {
        possiblyWriteQuizToDisk();
      }
    });
    
    switchDialog(QuizDialog.class.toString());
      
  }
  
  public void showFxDialog(final FxDialog dialog) {
    SwingUtilities.invokeLater(() -> {
      dialog.setSize(FRAME_SIZE);
      
      this.mainPanel.add(dialog, FxDialog.class.toString());
      
      switchDialog(FxDialog.class.toString());
    });
  }
  
  public void backFromFxDialog() {
    SwingUtilities.invokeLater(() -> {
      switchDialog(MainDialogFx.class.toString());
    });
  }
  
  private void updateIcuEnabledState() {
    if (this.actionMap != null && this.model.getIcuQuestions() != null) {
      
      if (this.actionMap.get(Actions.MC_INTENSIVE_CARE) != null) {
        
        final int count = this.model.getIcuQuestionsMc().getQuestionCount();
        
        this.actionMap.get(Actions.MC_INTENSIVE_CARE).setEnabled(count > 0);
        
        final String toolTip = "\"Sammelkorb\" für Fragen, die gezielt wiederholt werden sollen.\n" + 
            "Zurzeit befinden sich " + count + " Frage(n) aus den Themengebieten\n" +
            "und den Original-Überprüfungen in der Intensivstation.";
        this.actionMap.get(Actions.MC_INTENSIVE_CARE).putValue(Action.SHORT_DESCRIPTION, toolTip);
        
      }
      
      if (this.actionMap.get(Actions.TEXT_INTENSIVE_CARE) != null) {
        
        final int count = this.model.getIcuQuestionsText().getQuestionCount();
        
        this.actionMap.get(Actions.TEXT_INTENSIVE_CARE).setEnabled(count > 0);
        
        final String toolTip = "\"Sammelkorb\" für Fragen, die gezielt wiederholt werden sollen.\n" + 
            "Zurzeit befinden sich " + count + " Frage(n) aus den Themengebieten\n" +
            "und den Original-Überprüfungen in der Intensivstation.";
        this.actionMap.get(Actions.TEXT_INTENSIVE_CARE).putValue(Action.SHORT_DESCRIPTION, toolTip);
        
      }
      
    }
  }
  
  public void showPleaseWaitDialog(String message, SwingWorker worker) {
    showPleaseWaitDialog(message, worker, true);
  }
  
  public void showPleaseWaitDialogWithProgress(String message, SwingWorker worker) {
    showPleaseWaitDialog(message, worker, false);
  }
  
  private void showPleaseWaitDialog(String message, SwingWorker worker, final boolean indeterminate) {
    m_pleaseWaitDialog = new PleaseWaitDialog(this, message, worker, indeterminate);
    SwingUtilities.invokeLater(new Runnable() {
      public void run() {
        m_pleaseWaitDialog.setVisible(true);
      }
    });
  }

  public void showMessageDialog(final Object message, final int messageType) {
    final String title = TrainerTextResources.getString("APPLICATION_TITLE_SHORT");
    showMessageDialog(title, message, messageType, null, null);
  }
  
  public void showMessageDialog(final String title, final Object message, final int messageType) {
    showMessageDialog(title, message, messageType, null, null);
  }
  
  public void showMessageDialog(final String title, final Object message, final int messageType,
      final Runnable followUpAction) {
    showMessageDialog(title, message, messageType, null, followUpAction);
  }
  
  public void showMessageDialog(final String title, final Object message, final int messageType, final ImageIcon icon) {
    showMessageDialog(title, message, messageType, icon, null);
  }
  
  public void showMessageDialog(final String title, final Object message, final int messageType, final ImageIcon icon,
      final Runnable followUpAction) {
    TrainerMain.showMessageDialog(getRootPane(), title, message, messageType, icon, followUpAction);
  }

}
