package de.kreawi.trainer.db;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;

import javax.sql.DataSource;

import org.javalite.activejdbc.LazyList;

import de.kreawi.trainer.KreawiGlobals;
import de.kreawi.trainer.TrainerSettingsConstants;
import de.kreawi.trainer.TrainerStorage;

public class TrainerSettingsModel extends DbAccessModel {
  
  private List<SettingsChangedListener> settingsChangedListeners = new ArrayList<SettingsChangedListener>();
  private HashMap<String, String> defaultSettings;
  
  public TrainerSettingsModel(final DataSource dataSource) {
    super(dataSource);
    
    runWithDbConn(() -> {
      Setting.createTableIfNotExists();
    });
    
  }

  public String getValue(final String key) {
    return getValue(key, null);
  }
  
  public boolean isValueSet(final String key) {
    final String value = getValue(key, null);
    return value != null && !"".equals(value);
  }
  
  private String getValue(final String key, final String defaultIfNoDefault) {
    return callWithDbConn(() -> {
      final Setting s = Setting.findOrInit(Setting.DB_KEY, key);
      if (s.isNew()) {
        final HashMap<String, String> defaults = getDefaults();
        if (defaults.containsKey(key)) {
          s.setValue(defaults.get(key));
          s.saveIt();  // only save if default specified
        } else {
          return defaultIfNoDefault;
        }
      }
      return s.getValue();
    });
  }
  
  public void setValue(final String key, final String value) {
    runWithDbConn(() -> {
      final Setting s = Setting.findOrInit(Setting.DB_KEY, key);
      s.setValue(value);
      s.saveIt();
      notifySettingChangedListeners(key);
    });
  }
  
  public int getIntValue(final String key) {
    return Integer.parseInt(getValue(key, Integer.toString(0)));
  }
  
  public void setIntValue(final String key, final int value) {
    setValue(key, Integer.toString(value));
  }

  public long getLongValue(final String key) {
    return Long.parseLong(getValue(key, Long.toString(0l)));
  }
  
  public void setLongValue(final String key, final long value) {
    setValue(key, Long.toString(value));
  }
  
  public boolean getBooleanValue(final String key) {
    return Boolean.parseBoolean(getValue(key, Boolean.toString(false)));
  }
  
  public void setBooleanValue(final String key, final boolean value) {
    setValue(key, Boolean.toString(value));
  }
  
  private HashMap<String, String> getDefaults() {
    if (this.defaultSettings == null) {
      final HashMap<String, String> defaults = new HashMap<String, String>();
      defaults.put(TrainerSettingsConstants.KEY_JUSTIFY_TEXT, "true");
      defaults.put(TrainerSettingsConstants.KEY_START_MAXIMIZED, "false");
      defaults.put(TrainerSettingsConstants.KEY_UPDATE, "false");
      defaults.put(TrainerSettingsConstants.KEY_QUESTION_COMMENT_FONT_SIZE, "-1");
      defaults.put(TrainerSettingsConstants.KEY_ANSWERS_FONT_SIZE, "-1");
      defaults.put(TrainerSettingsConstants.KEY_EVALUATION_FONT_SIZE, "-1");
      defaults.put(TrainerSettingsConstants.KEY_MAX_CARD_BOX_LEVEL, "4");
      defaults.put(TrainerSettingsConstants.KEY_CARD_BOX_ACTIVE, "true");
      defaults.put(TrainerSettingsConstants.KEY_DEFAULT_CARD_BOX_LEVEL, "0");
      defaults.put(TrainerSettingsConstants.KEY_USE_UNANSWERED_FOR_CARD_BOX, "false");
      defaults.put(TrainerSettingsConstants.KEY_USE_EXTERNAL_IMAGE_VIEWER, "false");
      defaults.put(TrainerSettingsConstants.KEY_USE_EXTERNAL_AUDIO_PLAYER, "false");
      defaults.put(TrainerSettingsConstants.KEY_SHOW_CATEGORY_SELECTION_MC_EXPLANATION_MESSAGE, "true");
      defaults.put(TrainerSettingsConstants.KEY_SHOW_CATEGORY_SELECTION_TEXT_EXPLANATION_MESSAGE, "true");
      defaults.put(TrainerSettingsConstants.KEY_SHOW_ICU_EXPLANATION_MESSAGE, "true");
      defaults.put(TrainerSettingsConstants.KEY_SHOW_NOTES_EXPLANATION_MESSAGE, "true");
      defaults.put(TrainerSettingsConstants.KEY_SHOW_EXAM_SELECTION_EXPLANATION_MESSAGE, "true");
      defaults.put(TrainerSettingsConstants.KEY_SHOW_EXAM_MODE_EXPLANATION_MESSAGE, "true");
      defaults.put(TrainerSettingsConstants.KEY_SHOW_ICU_FILTER_EXPLANATION_MESSAGE, "true");
      defaults.put(TrainerSettingsConstants.KEY_SHOW_UNANSWERED_FILTER_EXPLANATION_MESSAGE, "true");
      defaults.put(TrainerSettingsConstants.KEY_SHOW_CARD_BOX_FILTER_EXPLANATION_MESSAGE, "true");
      defaults.put(TrainerSettingsConstants.KEY_SHOW_HISTORY_EXPLANATION_MESSAGE, "true");
      defaults.put(TrainerSettingsConstants.KEY_SHOW_ICU_FILTER_EXPLANATION_MESSAGE, "true");
      defaults.put(TrainerSettingsConstants.KEY_SHOW_TEXT_SEARCH_EXPLANATION_MESSAGE, "true");
      defaults.put(TrainerSettingsConstants.KEY_SHOW_TUTORIAL_ON_STARTUP, "true");
      this.defaultSettings = defaults;
    }
    return this.defaultSettings;
  }

  public Properties getSettingProperties() {
    return callWithDbConn(() -> {
      final SettingProperties props = new SettingProperties(TrainerStorage.loadOldPropertiesDefaults(), this);
      final LazyList<Setting> allSettings = Setting.findAll();
      for (Setting s : allSettings) {
        props.setPropertyWithoutDb(s.getKey(), s.getValue());
      }
      return props;
    });
  }

  public void addSettingChangedListener(final SettingsChangedListener listener) {
    this.settingsChangedListeners.add(listener);
  }

  public void removeSettingChangedListener(final SettingsChangedListener listener) {
    this.settingsChangedListeners.remove(listener);
  }
  
  public void notifyAllSettingsChangedListeners() {

    if (this.settingsChangedListeners == null || this.settingsChangedListeners.size() == 0) {
      return;
    }
    
    runWithDbConn(() -> {
      final LazyList<Setting> allSettings = Setting.findAll();
      for (Setting setting : allSettings) {
        notifySettingChangedListeners(setting.getKey());
      }
    });
  }

  private void notifySettingChangedListeners(final String key) {
    for (SettingsChangedListener listener : this.settingsChangedListeners) {
      listener.settingChanged(key);
    }
  }

  public interface SettingsChangedListener {
    public void settingChanged(final String key);
  }

}
