package de.kreawi.trainer.dialogs.onlineGame;

import java.io.File;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.kreawi.trainer.TrainerFrame;
import de.kreawi.trainer.dialogs.FxDialog;
import de.kreawi.trainer.onlinegame.OnlineGameMain;
import de.kreawi.trainer.onlinegame.OnlineGameModel.OnlineGameStatus;
import de.kreawi.trainer.onlinegame.OnlineGameSettings;
import de.kreawi.trainer.onlinegame.entities.OnlineGameFriendList;
import de.kreawi.trainer.onlinegame.entities.OnlineGameQuestionResult;
import de.kreawi.trainer.onlinegame.entities.OnlineGameQuiz;
import de.kreawi.trainer.onlinegame.entities.OnlineGameQuiz.QuizStatus;
import de.kreawi.trainer.onlinegame.entities.OnlineGameTopic;
import de.kreawi.trainer.onlinegame.entities.OnlineGameUser;
import de.kreawi.trainer.onlinegame.net.FriendListReceiver;
import de.kreawi.trainer.onlinegame.net.OnlineGameCommunicator.ServerFailureCallback;
import de.kreawi.trainer.onlinegame.net.OnlineGameQuizCommunicator.TopicWithQuestionsReceiver;
import de.kreawi.trainer.onlinegame.net.OnlineGameQuizCommunicator.TopicsReceiver;
import de.kreawi.trainer.util.PasswordGenerator;
import de.kreawi.trainer.util.TrainerHyperlinkListener;
import javafx.application.Platform;
import javafx.scene.Cursor;
import javafx.scene.layout.Pane;
import javafx.scene.layout.StackPane;
import javafx.stage.FileChooser;
import javafx.stage.FileChooser.ExtensionFilter;

public class OnlineGameContainer extends StackPane {
  
  private static final long ERROR_PAGE_DELAY = 3000l;

  private final OnlineGameMain onlineGame;
  
  private Dashboard dashboard;
  
  private final WaitPage waitPage;

  private FriendsList friendsListPage;

  private Pane externalPage;

  private final TrainerFrame trainerFrame;

  private final Runnable signupHelpAction;

  private static final Logger logger = LoggerFactory.getLogger(OnlineGameContainer.class); 
  
  public OnlineGameContainer(final OnlineGameMain onlineGame, final TrainerFrame trainerFrame,
      final Runnable signupHelpAction) {
    super();
    
    this.onlineGame = onlineGame;
    
    this.trainerFrame = trainerFrame;
    
    this.waitPage = new WaitPage(this);
    
    this.signupHelpAction = signupHelpAction;
    
    showFirstPage();
    
  }

  private void showFirstPage() {
    final OnlineGameSettings settings = this.onlineGame.getModel().getSettings();
    final String login = settings.getLogin();
    if (login == null || "".equals(login)) {
      showSignupPage();
    } else {
      login();
    }
  }

  public void showDashboard() {
    Platform.runLater(() -> {
      if (this.dashboard == null) {
        this.dashboard = new Dashboard(onlineGame, this);
      }
      showPage(this.dashboard);
    });
  }
  
  private void showSignupPage() {
    Platform.runLater(() -> {
      showPage(new Signup(this.onlineGame, this));
    });
  }

  public void signup(final String email, final String username) {
    
    showWaitPage();
    
    final Path profileImage = null;
    final String mobile = null;
    final Runnable onSuccess = () -> {
      showPage(new NotActivatedPage(OnlineGameContainer.this));
    };
    final ServerFailureCallback onFailure = new ServerFailureCallback() {
      @Override
      public void run(String serverMessage) {
        final String message = "Konnte diesen Benutzer nicht registrieren!" + 
            ((serverMessage != null && !"".equals(serverMessage))?"\nServer meldet:\n" + serverMessage:"");
        showErrorPage(message);
      }
    };
    final String password = PasswordGenerator.generateRandomPassword(14);
    this.onlineGame.signup(email, username, password, profileImage, mobile, onSuccess, onFailure);
    
  }
  
  public void login() {

    showWaitPage();
    
    final Runnable onSuccess = () -> {
      // TODO set over big button!!!
      this.onlineGame.setStatus(OnlineGameStatus.ONLINE, () -> {
        showDashboard();
      }, () -> {
        showDashboard();
      });
      
    };
    final ServerFailureCallback onFailure = (serverMessage) -> {
      showPage(new NotActivatedPage(serverMessage, OnlineGameContainer.this));
      showErrorPage("Konnte diesen Benutzer nicht anmelden!");
    };
    this.onlineGame.login(onSuccess, onFailure);
    
  }
  
  public void showEditProfilePage() {
    Platform.runLater(() -> {
      showPage(new EditProfile(this.onlineGame, this));
    });
  }
  
  public void editProfile(final String email, final String username) {
    final Runnable onSuccess = () -> {
      logger.info("Successfully upadeted profile data.");
      Platform.runLater(() -> {
        this.dashboard.reloadProfile();
        showDashboard();
      });
    };
    final ServerFailureCallback onFailure = (errorMessage) -> {
      logger.error("Error on upadeting profile data!");
      showDashboard();
      if (errorMessage == null || errorMessage.length() == 0) {
        showErrorPage("Fehler beim Aktualisieren des Profils!");
      } else {
        showErrorPage(errorMessage);
      }
    };
    showWaitPage();
    logger.info("Submitting new profile data...");
    
    final OnlineGameSettings settings = this.onlineGame.getModel().getSettings();
    final OnlineGameUser profileUser = settings.getProfileUser();
    this.onlineGame.editProfile(email, username, settings.getPassword(), null, profileUser.getMobile(), onSuccess,
        onFailure);
    
  }
  
  public void updateProfileImage() {
    final FileChooser fc = new FileChooser();
    fc.setTitle("Neues Profilbild wählen");
    fc.setInitialDirectory(new File(System.getProperty("user.home")));
    final ExtensionFilter imageFiles = new ExtensionFilter("Bilddateien", "*.jpg", "*.JPG", "*.jpeg", "*.JPEG",
        "*.png", "*.PNG", "*.gif", "*.GIF");
    final ExtensionFilter allFiles = new ExtensionFilter("Alle Dateien", "*.*");
    fc.getExtensionFilters().addAll(imageFiles, allFiles);
    fc.setSelectedExtensionFilter(imageFiles);
    final File choosenFile = fc.showOpenDialog(this.getScene().getWindow());
    if (choosenFile != null) {
      final Runnable onSuccess = () -> {
        logger.info("Successfully upadeted profile image.");
        Platform.runLater(() -> {
          this.dashboard.reloadProfile();
          showDashboard();
        });
      };
      final ServerFailureCallback onFailure = (errorMessage) -> {
        logger.error("Error on upadeting profile image!");
        showDashboard();
        if (errorMessage == null || errorMessage.length() == 0) {
          showErrorPage("Fehler beim Aktualisieren des Profilbildes!");
        } else {
          showErrorPage(errorMessage);
        }
      };
      showWaitPage();
      logger.info("Submitting new profile image...");
      
      final OnlineGameSettings settings = this.onlineGame.getModel().getSettings();
      final OnlineGameUser profileUser = settings.getProfileUser();
      this.onlineGame.editProfile(profileUser.getEmail(), profileUser.getUserName(), settings.getPassword(),
          choosenFile.toPath(), profileUser.getMobile(), onSuccess, onFailure);
    }
  }

  public void resetAllData() {
    this.onlineGame.resetAllData();
    showSignupPage();
  }
  
  public void showFriendsList() {
    
    showWaitPage();
    
    final FriendListReceiver onSuccess = new FriendListReceiver() {
      @Override
      public void run(final OnlineGameFriendList friendList) {
        final Comparator<OnlineGameUser> userListComparator = new Comparator<OnlineGameUser>() {
          @Override
          public int compare(OnlineGameUser ogu1, OnlineGameUser ogu2) {
            return ogu1.getUserName().compareTo(ogu2.getUserName());
          }
        };
        final List<OnlineGameUser> friendsList = Arrays.asList(friendList.getNormal());
        friendsList.sort(userListComparator);
        final Runnable arbitraryOpponentAction = () -> {
          showWaitPage();
          Platform.runLater(() -> {
              setCursor(Cursor.WAIT);
          });
          final List<OnlineGameUser> openUsersList = Arrays.asList(friendList.getOpen());
          openUsersList.sort(userListComparator);
          Platform.runLater(() -> {
            final OpenUsersList openUsersPage = new OpenUsersList(openUsersList, OnlineGameContainer.this);
            showPage(openUsersPage);
            setCursor(Cursor.DEFAULT);
          });
        };
        Platform.runLater(() -> {
          OnlineGameContainer.this.friendsListPage = new FriendsList(friendsList, arbitraryOpponentAction,
              OnlineGameContainer.this);
          showFriendsListNoReload();
        });
      }
    };
    final Runnable onFailure = () -> {showErrorPage("Spielerliste konnte nicht geladen werden!");};
    this.onlineGame.getFriendList(onSuccess, onFailure);
  }
  
  public void showFriendsListNoReload() {
    showPage(this.friendsListPage);
  }

  public void inviteUser(final OnlineGameUser user) {
    
    showWaitPage();
    
    final Runnable onSuccess = () -> {
      this.onlineGame.triggerNotificationCheck();
      showDashboard();
    };
    final Runnable onFailure = () -> {
      showErrorPage("Konnte Einladung nicht versenden!");
    };
    this.onlineGame.inviteUserForQuiz(user, onSuccess, onFailure);
    
  }

  public void removeFriend(final OnlineGameUser user) {
    
    showWaitPage();
    
    final Runnable onSuccess = () -> {
      showFriendsList();
    };
    final Runnable onFailure = () -> {
      showErrorPage("Konnte Spieler nicht von der Freunde-Liste entfernen!");
    };
    this.onlineGame.removeFriend(user, onSuccess, onFailure);
    
  }

  public void acceptInvitation(final OnlineGameQuiz quiz) {
    
    showWaitPage();
    
    quiz.setStatus(QuizStatus.SUBMITTING_DATA);

    final Runnable onSuccess = () -> {
      // Start quiz and select topic have two API calls, but follow each other directly
      // => should receive ACCEPTED directly after checking notifications
      this.onlineGame.triggerNotificationCheck();
      showDashboard();
    };
    final Runnable onFailure = () -> {
      showErrorPage("Die Einladung konnte nicht angenommen werden!");
    };
    this.onlineGame.acceptInvitation(quiz, onSuccess, onFailure);
  }

  public void rejectInvitation(final OnlineGameQuiz quiz) {
    
    showWaitPage();
    
    quiz.setStatus(QuizStatus.SUBMITTING_DATA);

    final Runnable onSuccess = () -> {
      showDashboard();
    };
    final Runnable onFailure = () -> {
      showErrorPage("Die Einladung konnte nicht abgelehnt werden!");
    };
    this.onlineGame.rejectInvitation(quiz, onSuccess, onFailure);
  }

  public void startQuiz(final OnlineGameQuiz quiz) {
    
    showWaitPage();
    
    quiz.setStatus(QuizStatus.SUBMITTING_DATA);

    final Runnable onSuccess = () -> {
      // Start quiz and select topic have two API calls, but follow each other directly
      // => should receive SELECT_TOPIC directly after checking notifications
      this.onlineGame.triggerNotificationCheck();
      showDashboard();
    };
    final Runnable onFailure = () -> {
      showErrorPage("Das Quiz konnte nicht gestartet werden!");
    };
    this.onlineGame.startQuiz(quiz, onSuccess, onFailure);
  }
  
  public void selectTopic(final OnlineGameQuiz quiz) {

    showWaitPage();
    
    final OnlineGameContainer ogc = this;
    final TopicsReceiver topicsReceiver = new TopicsReceiver() {
      @Override
      public void run(final List<OnlineGameTopic> topics) {
        Platform.runLater(() -> {
          showPage(new SelectTopic(quiz, topics, ogc));
        });
      }
    };
    
    final Runnable onFailure = () -> {
      showErrorPage("Die Themen konnten nicht geladen werden!");
    };
    
    this.onlineGame.loadTopics(quiz, topicsReceiver, onFailure);

  }
  
  /** Start round (if other player selected the topic) **/
  public void playRound(final OnlineGameQuiz quiz) {
    playRound(quiz, null);
  }
  
  /** Start round (if this player selected the topic) **/
  public void playRound(final OnlineGameQuiz quiz, final OnlineGameTopic topic) {

    showWaitPage();
    
    final OnlineGameContainer ogc = this;
    final TopicWithQuestionsReceiver questionsReceiver = (t) -> {
      
      Platform.runLater(() -> {
        final TrainerHyperlinkListener hyperlinkListener = new TrainerHyperlinkListener(
            OnlineGameContainer.this.trainerFrame);
        final PlayRound playRound = new PlayRound(quiz, t, onlineGame, ogc, hyperlinkListener);
        showExternalPage(playRound);
      });
    };
    
    final Runnable onFailure = () -> {
      showErrorPage("Keine Verbindung beim Laden der Fragen!");
    };

    
    if (topic == null) {
      this.onlineGame.loadQuestions(quiz, questionsReceiver, onFailure);
    } else {
      this.onlineGame.submitTopicLoadQuestions(quiz, topic, questionsReceiver, onFailure);
    }
  }

  public void submitRound(final OnlineGameQuiz quiz, final int roundNo, final String timeTaken, int rightQuestions,
      int wrongQuestions, final OnlineGameQuestionResult[] questionResults) {
    
    showWaitPage();   // this also returns from the external page
    
    quiz.setStatus(QuizStatus.SUBMITTING_DATA);
    
    final Runnable onFailure = () -> {
      showErrorPage("Keine Verbindung beim Übermitteln der Antworten!");
    };

    final Runnable onSuccess = () -> {
      
      this.onlineGame.triggerNotificationCheck();
      
      OnlineGameContainer.this.showQuizResults(quiz);
      
    };
    this.onlineGame.submitRound(quiz.getId(), roundNo, timeTaken.toString(), "" + rightQuestions,
        "" + wrongQuestions, questionResults, onSuccess, onFailure);
  }

  public void showQuizResults(final OnlineGameQuiz quiz) {

    showWaitPage();
    
    final OnlineGameContainer ogc = this;
    final Runnable onSuccess = () -> {
      Platform.runLater(() -> {
        final ResultsOverview resultsPage = new ResultsOverview(quiz, this.onlineGame, ogc);
        showPage(resultsPage);
      });
    };
    
    final Runnable onFailure = () -> {
      showDashboard();
      showErrorPage("Keine Verbindung beim Laden der Ergebnisse!");
    };

    this.onlineGame.loadQuizResults(quiz, onSuccess, onFailure);
    
  }

  public void showWaitPage() {
    
    if (this.externalPage != null) {
      backFromExternalPage();
    }
    
    if (!this.getChildren().contains(this.waitPage) || 
        this.getChildren().lastIndexOf(this.waitPage) != this.getChildren().size() - 1) {

      logger.info("Showing wait page layer.");
      
      Platform.runLater(() -> {
        this.getChildren().remove(this.waitPage);
        this.getChildren().add(this.waitPage);
      });
      
    } else {
      logger.info("Wait page already shown.");
    }
  }

  public void showErrorPage(final String message) {
    
    if (this.externalPage != null) {
      backFromExternalPage();
    }
    
    final ErrorPage ep = new ErrorPage(message);
    
    Platform.runLater(() -> {
      this.getChildren().remove(this.waitPage);
      this.getChildren().add(ep);  // no showPage() -> no clear!!!
    });
    
    new Timer().schedule(new TimerTask() {
      @Override
      public void run() {
        Platform.runLater(() -> {
          OnlineGameContainer.this.getChildren().remove(ep);
        });
      }
    }, ERROR_PAGE_DELAY);
  }

  private void showPage(final Pane page) {
    
    if (this.externalPage != null) {
      backFromExternalPage();
    }
    
    Platform.runLater(() -> {
      this.getChildren().remove(this.waitPage);
      this.getChildren().clear();  // TODO: transition???
      this.getChildren().add(page);
    });
  }

  private void showExternalPage(final Pane externalPage) {
    
    this.externalPage = externalPage;
    
    final FxDialog fxDialog = new FxDialog("kreawi-Duell", externalPage);
    
    this.trainerFrame.showFxDialog(fxDialog);
  }
  
  public void backFromExternalPage() {
    this.externalPage = null;
    this.trainerFrame.backFromFxDialog();
  }

  public Runnable getSignupHelpAction() {
    return signupHelpAction;
  }

}
