package de.kreawi.trainer.dialogs.onlineGame.custom_nodes;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.kreawi.trainer.dialogs.onlineGame.OnlineGameContainer;
import de.kreawi.trainer.onlinegame.OnlineGameMain;
import de.kreawi.trainer.onlinegame.OnlineGameMain.OnlineStatusChangedListener;
import de.kreawi.trainer.onlinegame.OnlineGameModel.OnlineGameStatus;
import de.kreawi.trainer.onlinegame.OnlineGameSettings;
import de.kreawi.trainer.onlinegame.entities.OnlineGameUser;
import de.kreawi.trainer.util.TrainerTextResources;
import javafx.application.Platform;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.geometry.Insets;
import javafx.geometry.Pos;
import javafx.scene.Cursor;
import javafx.scene.control.Alert;
import javafx.scene.control.Alert.AlertType;
import javafx.scene.control.ContextMenu;
import javafx.scene.control.RadioMenuItem;
import javafx.scene.control.ToggleGroup;
import javafx.scene.control.Tooltip;
import javafx.scene.input.MouseButton;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.BorderPane;
import javafx.scene.shape.Circle;
import javafx.scene.text.Text;

public class ProfileHeader extends BorderPane {
  
  @SuppressWarnings("unused")
  private static final Logger logger = LoggerFactory.getLogger(ProfileHeader.class);
  private final OnlineGameMain onlineGame;
  private final OnlineGameContainer onlineGameContainer;
  private BorderPane captionPane = new BorderPane(); 

  public ProfileHeader(final OnlineGameMain onlineGame, final OnlineGameContainer onlineGameContainer) {
    
    this.onlineGame = onlineGame;
    this.onlineGameContainer = onlineGameContainer;
    
    getStylesheets().add("/de/kreawi/trainer/dialogs/onlineGame/online_game.css");
    
    getStyleClass().add("profile-header");
    
    loadProfile();
    
    setCenter(captionPane);
    
    this.onlineGame.addStatusChangedListener(new OnlineStatusChangedListener() {
      @Override
      public void onStatusChanged(OnlineGameStatus oldStatus, OnlineGameStatus newStatus) {
        Platform.runLater(() -> {
          loadOnlineStatusButton();
        });
      }
    });
    
    loadOnlineStatusButton();
  }

  public void loadProfile() {
    final OnlineGameSettings settings = this.onlineGame.getModel().getSettings();
    final OnlineGameUser profileUser = settings.getProfileUser();
    
    final ProfileImage profileImage = new ProfileImage(profileUser, true);
    profileImage.setCursor(Cursor.HAND);
    profileImage.setOnMouseClicked(new EventHandler<MouseEvent>() {
      @Override
      public void handle(MouseEvent event) {
        if (!event.getButton().equals(MouseButton.PRIMARY)) {
          return;
        }
        ProfileHeader.this.onlineGameContainer.updateProfileImage();
      }
    });
    setLeft(profileImage);
    
    final Text caption = new Text(profileUser.getUserName());
    BorderPane.setAlignment(caption, Pos.CENTER_LEFT);
    BorderPane.setMargin(caption, new Insets(0, 0, 0, 15));
    caption.setCursor(Cursor.HAND);
    caption.setOnMouseClicked(new EventHandler<MouseEvent>() {
      @Override
      public void handle(MouseEvent event) {
        if (!event.getButton().equals(MouseButton.PRIMARY)) {
          return;
        }
        ProfileHeader.this.onlineGameContainer.showEditProfilePage();
      }
    });
    
    this.captionPane.setLeft(caption);
    
  }
  
  private void loadOnlineStatusButton() {
    final double onlineStatusButtonRadius = 7d;
    final OnlineGameStatus currentOnlineStatus = this.onlineGame.getStatus();
    final String styleClass;
    final Circle c = new Circle(onlineStatusButtonRadius);
    final String statusText;
    if (OnlineGameStatus.ONLINE.equals(currentOnlineStatus)) {
      statusText = "ONLINE";
      styleClass = "green";
    } else if (OnlineGameStatus.BUSY.equals(currentOnlineStatus)) {
      statusText = "BESCHÄFFTIGT";
      styleClass = "yellow";
    } else {
      statusText = "OFFLINE";
      styleClass = "red";
    }
    c.getStyleClass().add(styleClass);
    c.setCursor(Cursor.HAND);
    c.setOnMouseClicked(new EventHandler<MouseEvent>() {
      @Override
      public void handle(MouseEvent event) {
        if (ProfileHeader.this.onlineGame.isLoggedIn()) {
          final OnlineGameStatus status = ProfileHeader.this.onlineGame.getStatus();
          final ContextMenu contextMenu = createOnlineStatusContextMenu(status);
          contextMenu.show(c, event.getScreenX(), event.getScreenY());
        } else {
          final String message = "Bitte schließen Sie erst die Registrierung des kreawi-Duell-Benutzers ab!";
          final Alert alert = new Alert(AlertType.WARNING, message);
          alert.setTitle(TrainerTextResources.getString("APPLICATION_TITLE"));
          alert.show();
        }
      }
    });
    Tooltip tootip = new Tooltip("Status: " + statusText + "\nKlicken Sie hier um den Status zu ändern!");
    Tooltip.install(c, tootip);

    BorderPane.setAlignment(c, Pos.CENTER_LEFT);
    BorderPane.setMargin(c, new Insets(10));
    this.captionPane.setCenter(c);
  }
  
  protected ContextMenu createOnlineStatusContextMenu(final OnlineGameStatus currentStatus) {

    final Runnable onSuccess = () -> {
      logger.info("Successfully changed online status!");
    };
    final Runnable onFailure = () -> {
      logger.error("Failed to change online status!");
    };
    
    ToggleGroup group = new ToggleGroup();
    
    final RadioMenuItem radioSetOnline = new RadioMenuItem("ONLINE: Offen für Einladungen");
    radioSetOnline.setOnAction(new EventHandler<ActionEvent>() {
      @Override
      public void handle(ActionEvent event) {
        final OnlineGameStatus newStatus = OnlineGameStatus.ONLINE;
        onlineGame.setStatus(newStatus, onSuccess, onFailure);
      }
    });
    radioSetOnline.setToggleGroup(group);
    radioSetOnline.setSelected(OnlineGameStatus.ONLINE.equals(currentStatus));
    
    final RadioMenuItem radioSetBusy = new RadioMenuItem("BESCHÄFFTIGT: Keine neuen Einladungen");
    radioSetBusy.setOnAction(new EventHandler<ActionEvent>() {
      @Override
      public void handle(ActionEvent event) {
        final OnlineGameStatus newStatus = OnlineGameStatus.BUSY;
        onlineGame.setStatus(newStatus, onSuccess, onFailure);
      }
    });
    radioSetBusy.setToggleGroup(group);
    radioSetBusy.setSelected(OnlineGameStatus.BUSY.equals(currentStatus));
    
    final RadioMenuItem radioSetOffline = new RadioMenuItem("OFFLINE: Alle Spiele pausieren");
    radioSetOffline.setOnAction(new EventHandler<ActionEvent>() {
      @Override
      public void handle(ActionEvent event) {
        final OnlineGameStatus newStatus = OnlineGameStatus.OFFLINE;
        onlineGame.setStatus(newStatus, onSuccess, onFailure);
      }
    });
    radioSetOffline.setToggleGroup(group);
    radioSetOffline.setSelected(OnlineGameStatus.OFFLINE.equals(currentStatus));
    
    final ContextMenu cm = new ContextMenu();
    cm.getItems().addAll(radioSetOnline, radioSetBusy, radioSetOffline);
    
    return cm;
  }

}
