package de.kreawi.trainer.onlinegame.net;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.fasterxml.jackson.databind.ObjectMapper;

import de.kreawi.trainer.onlinegame.OnlineGameModel;
import de.kreawi.trainer.onlinegame.entities.OnlineGameResults;
import de.kreawi.trainer.onlinegame.entities.OnlineGameRoundScore;
import de.kreawi.trainer.onlinegame.entities.OnlineGameTopic;

public class OnlineGameQuizCommunicator extends OnlineGameCommunicator {

  private static final Logger logger = LoggerFactory.getLogger(OnlineGameQuizCommunicator.class);
  
  public OnlineGameQuizCommunicator(final OnlineGameModel model) {
    super(model);
  }

  public void inviteUserForQuiz(final String userId, final String friendId, final Runnable onSuccess,
      final Runnable onFailure) {
    
    final InviteData inviteData = new InviteData(userId, friendId);
    
    try {
      
      submitJsonData("/inviteUserForQuiz", (new ObjectMapper()).writeValueAsString(inviteData), null);
      
      onSuccess.run();
      
    } catch (Exception e) {
      logger.error("Error on inviting user for quiz!", e);
      onFailure.run();
    }
    
  }
  
  public void acceptQuizInvitation(final String userId, final String notificationId, final Runnable onSuccess,
      final Runnable onFailure) {
    
    acceptOrRejectQuizInvitation(userId, notificationId, AcceptRejectData.ACCEPT, onSuccess, onFailure);
    
  }
  
  public void rejectQuizInvitation(final String userId, final String notificationId, final Runnable onSuccess,
      final Runnable onFailure) {

    acceptOrRejectQuizInvitation(userId, notificationId, AcceptRejectData.REJECT, onSuccess, onFailure);
    
  }
  
  private void acceptOrRejectQuizInvitation(final String userId, final String notificationId,
      final String action, final Runnable onSuccess, final Runnable onFailure) {
    
    final AcceptRejectData acceptRejectData = new AcceptRejectData(userId, notificationId, action);

    try {
      
      submitJsonData("/acceptQuizInvitation", (new ObjectMapper()).writeValueAsString(acceptRejectData), null);
      
      onSuccess.run();
      
    } catch (Exception e) {
      logger.error("Error on accepting / rejecting quiz invitation!", e);
      onFailure.run();
    }
    
  }
  
  public void startQuiz(final String userId, final String quizId, final Runnable onSuccess,
      final Runnable onFailure) {

    final StartQuizData startQuizData = new StartQuizData(userId, quizId);

    try {
      
      submitJsonData("/startQuiz", (new ObjectMapper()).writeValueAsString(startQuizData), null);
      
      onSuccess.run();
      
    } catch (Exception e) {
      logger.error("Error on starting quiz!", e);
      onFailure.run();
    }
    
  }
  
  public void getTopics(final String quizId, final String userId,
      final TopicsReceiver onSuccess, final Runnable onFailure) {
    
    final HashMap<String, String> parameters = new HashMap<String, String>();
    parameters.put("recordType", RECORD_TYPE_QUIZ);
    
    try {
      
      final OnlineGameTopic[] topicsArray = receiveJsonData("/getQuizTopics", parameters, OnlineGameTopic[].class);
      
      final List<OnlineGameTopic> topics = Arrays.asList(topicsArray);
      
      onSuccess.run(topics);
      
    } catch (Exception e) {

      logger.error("Error while receiving topics!", e);
      
      onFailure.run();
      
    }
    
  }
  
  public void submitTopicGetQuestions(final String userId, final String quizId, final String topicId,
      final TopicWithQuestionsReceiver onSuccess, final Runnable onFailure) {
    
    final SubmitTopicData submitTopicData = new SubmitTopicData(userId, quizId, topicId);

    try {
      
      final OnlineGameTopic topic = submitJsonData("/getQuizTopicQuestions",
          (new ObjectMapper()).writeValueAsString(submitTopicData), OnlineGameTopic.class);
      
      onSuccess.run(topic);
      
    } catch (Exception e) {
      logger.error("Error on submitting topic and receiving questions!", e);
      onFailure.run();
    }
    
  }
  
  public void getQuestions(final String userId, final String quizId, final TopicWithQuestionsReceiver onSuccess,
      final Runnable onFailure) {
    
    final HashMap<String, String> parameters = new HashMap<String, String>();
    parameters.put("userId", userId);
    parameters.put("quizId", quizId);
    
    try {
      
      final OnlineGameTopic topic = receiveJsonData("/getQuizQuestions", parameters, OnlineGameTopic.class);
      
      onSuccess.run(topic);
      
    } catch (Exception e) {

      logger.error("Error while receiving questions!", e);
      
      onFailure.run();
      
    }
    
  }
  
  public void submitRoundScore(final OnlineGameRoundScore roundScore, final Runnable onSuccess, final Runnable onFailure) {
    
    try {
      
      submitJsonData("/submitQuizRoundScore", (new ObjectMapper()).writeValueAsString(roundScore), null);
      
      onSuccess.run();
      
    } catch (Exception e) {
      logger.error("Error on submitting round score!", e);
      onFailure.run();
    }
    
  }
  
  public void getQuizResults(final String userId, final String quizId, final ResultsReceiver onSuccess,
      final Runnable onFailure) {
    
    final HashMap<String, String> parameters = new HashMap<String, String>();
    parameters.put("userId", userId);
    parameters.put("quizId", quizId);
    
    try {
      
      final OnlineGameResults results = receiveJsonData("/getQuizThreadDetail", parameters, OnlineGameResults.class);
      
      onSuccess.run(results);
      
    } catch (Exception e) {

      logger.error("Error while receiving quiz results!", e);
      
      onFailure.run();
      
    }
        
  }

  public static class InviteData {
    
    private String userId;
    
    private String friendId;

    public InviteData(String userId, String friendId) {
      super();
      this.userId = userId;
      this.friendId = friendId;
    }

    public String getUserId() {
      return userId;
    }

    public void setUserId(String userId) {
      this.userId = userId;
    }

    public String getFriendId() {
      return friendId;
    }

    public void setFriendId(String friendId) {
      this.friendId = friendId;
    }
    
  }
  
  public static class AcceptRejectData {
    
    public final static String ACCEPT = "accept";
    
    public final static String REJECT = "reject";
    
    private String userId;
    
    private String notificationId;
    
    private String action;

    public AcceptRejectData(String userId, String notificationId, String action) {
      super();
      this.userId = userId;
      this.notificationId = notificationId;
      this.action = action;
    }

    public String getUserId() {
      return userId;
    }

    public void setUserId(String userId) {
      this.userId = userId;
    }

    public String getNotificationId() {
      return notificationId;
    }

    public void setNotificationId(String notificationId) {
      this.notificationId = notificationId;
    }

    public String getAction() {
      return action;
    }

    public void setAction(String action) {
      this.action = action;
    }
    
  }
  
  public static class StartQuizData {
    
    private String userId;

    private String quizChallengeId;
    
    public StartQuizData(final String userId, final String quizChallengeId) {
      super();
      this.userId = userId;
      this.quizChallengeId = quizChallengeId;
    }

    public String getUserId() {
      return userId;
    }

    public void setUserId(String userId) {
      this.userId = userId;
    }
    
    public String getQuizChallengeId() {
      return quizChallengeId;
    }

    public void setQuizChallengeId(String quizChallengeId) {
      this.quizChallengeId = quizChallengeId;
    }

  }
  
  public static class SubmitTopicData {
    private String userId;
    
    private String quizChallengeId;
    
    private String topicId;

    public SubmitTopicData(String userId, String quizId, String topicId) {
      super();
      this.userId = userId;
      this.quizChallengeId = quizId;
      this.topicId = topicId;
    }

    public String getUserId() {
      return userId;
    }

    public void setUserId(String userId) {
      this.userId = userId;
    }

    public String getQuizChallengeId() {
      return quizChallengeId;
    }

    public void setQuizChallengeId(String quizId) {
      this.quizChallengeId = quizId;
    }

    public String getTopicId() {
      return topicId;
    }

    public void setTopicId(String topicId) {
      this.topicId = topicId;
    }
    
  }
  
  public static interface TopicsReceiver {
    public void run(final List<OnlineGameTopic> topics);
  }
  
  public static interface TopicWithQuestionsReceiver {
    public void run(final OnlineGameTopic topic);
  }

  public static interface ResultsReceiver {
    public void run(final OnlineGameResults results);
  }
}
