package de.kreawi.trainer.update;

import java.awt.event.ActionEvent;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.URISyntaxException;
import java.nio.file.Files;
import java.nio.file.StandardCopyOption;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.ExecutionException;
import java.util.jar.Attributes;
import java.util.jar.JarFile;
import java.util.jar.JarInputStream;
import java.util.jar.Manifest;

import javax.swing.AbstractAction;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;
import javax.swing.filechooser.FileFilter;

import org.apache.commons.io.FileUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.kreawi.trainer.KreawiGlobals;
import de.kreawi.trainer.license.KreawiLicenseHandler;
import de.kreawi.trainer.license.RegistrationHandler;
import de.kreawi.trainer.update.autoupdate.AutoUpdater;
import de.kreawi.trainer.util.OSDetector;
import de.kreawi.trainer.util.TrainerTextResources;
import de.kreawi.trainer.util.swingx.PleaseWaitDialog;
import de.martinknauer.util.Util;
import de.martinknauer.util.Version;
import de.martinknauer.util.zip.ZipUtil;

public class Updater {

	public final static String FILE_NAME_TRAINER_JAR = "trainer.jar";

	public final static String FILE_NAME_UPDATE_ZIP = "trainerAutoupdate.zip";
	
	public final static String DIR_NAME_LIB = "lib";
	
	public final static String FILE_PATH_BUNDLE_MAIN = "Contents/";
	
	public final static String FILE_PATH_BUNDLE_JAVA = "Java/";  // since Trainer 2016 (integrated Java)

	public final static String FILE_PATH_BUNDLE_RESOURCES = "Resources/";  // since Trainer 2016 (integrated Java)

  public final static String APP_BUNDLE_EXTENSION = ".app";
  
	private final static String DIR_NAME_DEFAULT_DIR_WINDOWS = "kreawi Prüfungstrainer";
  private final static String DIR_NAME_DEFAULT_DIR_MAC = "KreawiTrainer" + KreawiGlobals.CURRENT_YEAR + APP_BUNDLE_EXTENSION;
	
	public final static String MIN_VERSION = "1.25.0.0";
	
	public final static String MAX_VERSION = "1.25.99.0";

	private Thread thread;

	private File workDir = new File(System.getProperty("user.dir"));

	private AbstractAction exitAction;

	private UpdaterFrame gui;

	private boolean updateSuccess = false;

  // Don't init logger before setting trainerAppDir for logging config!
	private static Logger logger; 

	public Updater() {
		this(false);
	}
	
	@SuppressWarnings("serial")
	public Updater(boolean silent) {
		this.exitAction = new AbstractAction(TrainerTextResources
				.getString("BUTTON_CLOSE")) {
			public void actionPerformed(ActionEvent e) {
				exit();
			}
		};
		exitAction.setEnabled(false);

		this.gui = new UpdaterFrame(TrainerTextResources.getString("UPDATE"),
				exitAction);

		gui.setVisible(!silent);
	}
	
	private boolean hasUpdatePermissions() {

		return Util.canWriteInDirectory(this.workDir);

	}

	private boolean canUpdateInWorkingDirectory() {
		final File trainerJar = getTrainerJar();
		return trainerJar.exists();
	}

  private synchronized void doUpdate(final boolean updateFullVersionData) {
    
		if (!hasUpdatePermissions()) {
			gui.appendMessage("\n"
					+ TrainerTextResources.getString("UPDATER_NO_PERMISSION")
					+ "\n");
			exitAction.setEnabled(true);
			return;
		}

		if (!isMinimumTrainerVersion()) {
      gui.appendMessage("\n" + TrainerTextResources.getStringWithSubstitution("UPDATER_MIN_VERSION_ERROR",
          KreawiGlobals.CURRENT_YEAR) + "\n");
      gui.appendMessage("\nInstalliert: " + getInstalledVersion().toNumericalString());
      gui.appendMessage("\nMin: " + MIN_VERSION);
			exitAction.setEnabled(true);
			return;
		}
		
		if (isAboveMaxTrainerVersion()) {
      gui.appendMessage("\n" + TrainerTextResources.getString("UPDATER_MAX_VERSION_ERROR") + "\n");
			gui.appendMessage("\nInstalliert: " + getInstalledVersion().toNumericalString());
			gui.appendMessage("\nMax: " + MAX_VERSION);
			exitAction.setEnabled(true);
			return;
		}
		
		if (!KreawiLicenseHandler.getInstance().hasValidLicence() || RegistrationHandler.isLicenseLocked()) {
		  gui.appendMessage(TrainerTextResources.getString("UPDATER_LICENSE_ERROR"));
      exitAction.setEnabled(true);
      return;
		}
		
		if (!isFullVersionUpdate() && !getDataFile().exists()) {
      final String fullVersionUpdateUrl = TrainerTextResources.getStringWithSubstitution(
          "LATEST_VERSION_DOWNLOAD_URL_FULL_VERSION",
          new String[] { KreawiGlobals.CURRENT_YEAR, KreawiGlobals.CURRENT_YEAR });
      gui.appendMessage(TrainerTextResources.getStringWithSubstitution("UPDATER_NO_FULL_VERSION_ERROR", fullVersionUpdateUrl));
      exitAction.setEnabled(true);
      return;
		}

		final File updateFile = getUpdateFile();

		final ZipUtil unzip = new ZipUtil();

		unzip.addProgressListener(gui);

		this.thread = new Thread(new Runnable() {
			public void run() {
				
				boolean deleteUpdateFile = false;
				if (!updateFile.exists()) {
				  logger.info("Data for update not found. (search location: " + updateFile.getAbsolutePath() + ")");
					try {
						/* file has to be loaded from the net */
						gui.appendMessage(TrainerTextResources.getString("UPDATE_DOWNLOAD_START") + "\n");
						downloadUpdate(updateFullVersionData);
						gui.appendMessage(TrainerTextResources.getString("UPDATE_DOWNLOAD_SUCCESS") + "\n\n");
						deleteUpdateFile = true;
					} catch (IOException e) {
					  logger.error("Unable to download update file", e);
						
						final String downloadUrl;
            if (updateFullVersionData) {
              downloadUrl = TrainerTextResources.getStringWithSubstitution("LATEST_VERSION_DOWNLOAD_URL_FULL_VERSION",
                  KreawiGlobals.CURRENT_YEAR);
            } else {
              downloadUrl = TrainerTextResources.getStringWithSubstitution("LATEST_VERSION_DOWNLOAD_URL",
                  KreawiGlobals.CURRENT_YEAR);
            }
            final String updateFailedText = TrainerTextResources.
              getStringWithSubstitution("AUTO_UPDATE_FAILED_TEXT", downloadUrl);
            
            JOptionPane.showMessageDialog(null, updateFailedText,
                TrainerTextResources.getString("AUTO_UPDATE_FAILED_TITLE"), JOptionPane.WARNING_MESSAGE);
						gui.appendMessage("\n" + TrainerTextResources.getString("UPDATE_DOWNLOAD_FAILED") + "\n\n");
						exitAction.setEnabled(true);
						return;
					} catch (Exception e) {
						gui.appendMessage(e.getMessage());
					}
				}

				try {
					gui.appendMessage("Installiere nach " + workDir + "\n");
					updateSuccess  = unzip.unZipFile(updateFile, workDir);
					if (OSDetector.getOs().isMac()) {
					  Files.move(new File(workDir, FILE_NAME_TRAINER_JAR).toPath(), new File(workDir.getParentFile(), FILE_PATH_BUNDLE_JAVA + FILE_NAME_TRAINER_JAR).toPath(), StandardCopyOption.REPLACE_EXISTING);
					  final File libDirSrc = new File(workDir, DIR_NAME_LIB);
					  if (libDirSrc.exists()) {
					    final File libDirTarget = new File(new File(workDir.getParentFile(), FILE_PATH_BUNDLE_JAVA), DIR_NAME_LIB);
					    FileUtils.copyDirectory(libDirSrc, libDirTarget);
					    FileUtils.deleteDirectory(libDirSrc);
					  }
					}
				} catch (IOException e) {
					deleteUpdateFile = true;
				}

				if (updateSuccess) {
					gui.appendMessage("\n" + TrainerTextResources.getString("UPDATER_SUCCESS"));
				} else {
					gui.appendMessage("\n" + TrainerTextResources.getString("UPDATER_ERROR"));
				}
				
				if (deleteUpdateFile) {
					updateFile.delete();
				}

				exitAction.setEnabled(true);
				
			}
		});

		this.thread.start();

	}

  public static File getUpdateFile() {
    return new File(new File(System.getProperty("user.dir")), FILE_NAME_UPDATE_ZIP);
  }
	
	private void downloadUpdate(final boolean updateFullVersionData) throws IOException {
		AutoUpdater au = new AutoUpdater(updateFullVersionData);
		au.retrieveUpdateInfos();
		if (AutoUpdater.hasAutoUpdatePermissions()) {
		  logger.info("Downloading update files from " + au.getUpdateUrl());
			au.downloadUpdate(gui);
		}
	}

  private File downloadUpdater(final boolean updateFullVersionData) throws IOException {
		AutoUpdater au = new AutoUpdater(updateFullVersionData);
		au.retrieveUpdateInfos();
		if (AutoUpdater.hasAutoUpdatePermissions()) {
			return au.performAutoupdaterUpdate();
		}
		return null;
	}
	
	private boolean isMinimumTrainerVersion() {

		Version installedVersion = getInstalledVersion();

		if (installedVersion == null) {
			return false;
		}

		Version minVersion = new Version(MIN_VERSION);

		return installedVersion.compareTo(minVersion) >= 0;

	}
	
	private boolean isAboveMaxTrainerVersion() {
		Version installedVersion = getInstalledVersion();

		if (installedVersion == null) {
			return true;
		}

		Version maxVersion = new Version(MAX_VERSION);

		return installedVersion.compareTo(maxVersion) > 0;
	}
	
	private Version getInstalledVersion() {
		
	  try (final JarFile trainerJar = new JarFile(getTrainerJar())) {
	    
			Version installedVersion = new Version(trainerJar.getManifest()
					.getMainAttributes().getValue(
							Attributes.Name.IMPLEMENTATION_VERSION));

			return installedVersion;
			
		} catch (IOException e) {
		  logger.error("Unable to retrieve version from trainer.jar!", e);
			return null;
		}
	}

  private File getTrainerJar() {
    final File trainerJar;
    if (OSDetector.getOs().isMac()) {
      trainerJar = new File(workDir.getParentFile(), FILE_PATH_BUNDLE_JAVA + FILE_NAME_TRAINER_JAR);
    } else {
      trainerJar = new File(workDir, FILE_NAME_TRAINER_JAR);
    }
    return trainerJar;
  }
  
  private File getDataFile() {
    if (OSDetector.getOs().isMac()) {
      return new File(
          new File(new File(workDir.getParentFile(), FILE_PATH_BUNDLE_RESOURCES), KreawiGlobals.QUESTION_DIR_NAME),
          KreawiGlobals.QUESTION_FILE_NAME);
    } else {
      return new File(new File(workDir, KreawiGlobals.QUESTION_DIR_NAME), KreawiGlobals.QUESTION_FILE_NAME);
    }
  }
	
	private boolean canUpdateBySettingsEntry() {
	  final String installDir = KreawiGlobals.getInstallDir();
	  if (installDir == null) {
	    logger.error("Cannot update by settings entry, because the 'install_dir'-entry " +
	        "does not exist in the user's settings file!");
	    return false;
	  }
	  
	  this.workDir = new File(installDir);
	  return canUpdateInWorkingDirectory();
	}
	
  private boolean canUpdateInDefaultDirectory() {

    if (OSDetector.getOs().isWindows()) {
      this.workDir = new File(System.getenv("PROGRAMFILES"), DIR_NAME_DEFAULT_DIR_WINDOWS);
    } else if (OSDetector.getOs().isMac()) {
      this.workDir = new File("/Applications/", DIR_NAME_DEFAULT_DIR_MAC);
      this.workDir = new File(new File(this.workDir, FILE_PATH_BUNDLE_MAIN), FILE_PATH_BUNDLE_RESOURCES);
    } else if (OSDetector.getOs().isLinux()) {
      this.workDir = new File(System.getProperty("user.home"), "kreawiTrainer");
    }

    return canUpdateInWorkingDirectory();

  }
	
	private boolean canUpdateByFileChooser() {
		this.gui.setEnabled(false);
		JOptionPane.showMessageDialog(this.gui, TrainerTextResources
				.getString("UPDATER_CHOOSE_DIR_MESSAGE"), TrainerTextResources
				.getString("UPDATER_CHOOSE_DIR_TITLE"), JOptionPane.INFORMATION_MESSAGE);
		
		logger.info("Opening File-Chooser...");

    JFileChooser fc = new JFileChooser();
		
		if (OSDetector.getOs().isMac()) {
			/* platform is mac os x
			 * we have to take care of mac os application bundles
			 */
		  logger.info("Mac OS X platform detected!");
		  
			fc.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
			fc.addChoosableFileFilter(new FileFilter() {

				@Override
				public boolean accept(File f) {
					if (f.getName().endsWith(APP_BUNDLE_EXTENSION)) {
            return new File(f, FILE_PATH_BUNDLE_MAIN + FILE_PATH_BUNDLE_JAVA + FILE_NAME_TRAINER_JAR).exists();
					} else {
						return f.isDirectory();
					}
				}

				@Override
				public String getDescription() {
					return TrainerTextResources.getString("APPLICATION_TITLE_SHORT");
				}

			});
		} else {
			/* non-mac os */
			fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
		}
		
		int answer = fc.showOpenDialog(this.gui);
		this.gui.setEnabled(true);
		this.gui.toFront();
		
		if (answer == JFileChooser.APPROVE_OPTION) {

			this.workDir = fc.getSelectedFile();
			
			if (OSDetector.getOs().isMac()) {
				/* platform is mac os x
				 * once again, we have to take care of mac os application
				 * bundles
				 */
				
				if (this.workDir.getName().endsWith(APP_BUNDLE_EXTENSION)) {
          this.workDir = new File(this.workDir,
              FILE_PATH_BUNDLE_MAIN + FILE_PATH_BUNDLE_RESOURCES);
				}
			}

			logger.info("Selected directory '" + this.workDir + "' for update");

		} else {
			return false;
		}
		
		return canUpdateInWorkingDirectory();
	}
	
	public void exit() {
	  if (this.thread == null || !this.thread.isAlive()) {
		  logger.info("Exiting updater.");
		  System.exit(0);
		} else {

			try {
				this.thread.join(20000);
			} catch (InterruptedException e) {
				/* do nothing */
			}

			if (!this.thread.isAlive()) {
			  logger.info("Exiting updater.");
				System.exit(0);
			} else {
			  logger.error("Exiting updater - error code 1.");
				System.exit(1);
			}

		}
	}

	public static void main(String[] args) {
	  
    // For Log4j configuration
    final File appDataDir = KreawiGlobals.getApplicationDirectory();
    System.setProperty("trainerAppDir", appDataDir.getAbsolutePath());
    
    logger = LoggerFactory.getLogger(Updater.class);

    // Old logger necessary because of ZipUtil (its using the logger!!!)
    de.martinknauer.util.logging.Logger.createInstance(false, appDataDir);

    final Updater updater = new Updater();

		KreawiGlobals.configureDefaultUI();
		final StringBuilder callArgs = new StringBuilder();
		for (String a:args) {
		  if (callArgs.length() > 0)
		    callArgs.append(" ");
		  callArgs.append(a);
		}
		logger.info("Updater ausgeführt mit diesen Argumenten: " + callArgs.toString());
		
    final boolean updateFullVersionData = KreawiGlobals.isFullVersionUpdate();
    
    String argDir = "";
    if (args.length > 0) {

			List<String> argList = Arrays.asList(args);

			if (argList.contains("-dev")) {
				KreawiGlobals.setDevMode(true);
			}

			if (argList.contains("-loadUpdater")) {
				loadUpdater(updateFullVersionData);
				return;
			}
			
			if (argList.contains("-instdir")) {
			  final int i = argList.indexOf("-instdir");
			  try {
			    final String instdir = argList.get(i + 1);
			    if (instdir != null && instdir.length() > 0) {
			      argDir = instdir;
			    }
			  } catch (IndexOutOfBoundsException e) {
			    logger.info("Verzeichnis-Parameter nicht gefunden!");
			  }
	    }
		}

		System.getProperties().put("apple.laf.useScreenMenuBar", "true");

		logger.info("Updater version " + (Util.getVersion() == null ? "N/A" : Util.getVersion()));

		if (updater.canUpdateInArgDirectory(argDir)) {
		  logger.info("Found trainer in passed in directory ('" + updater.workDir + "')...");
      updater.doUpdate(updateFullVersionData);
		} else if (updater.canUpdateInWorkingDirectory()) {
		  logger.info("Found trainer in working directory ('" + updater.workDir + "')...");
			updater.doUpdate(updateFullVersionData);
		} else if (updater.canUpdateBySettingsEntry()) {
		  logger.info("Found trainer by entry in settings file ('" + updater.workDir + "')...");
			updater.doUpdate(updateFullVersionData);
		} else if (updater.canUpdateInDefaultDirectory()) {
		  logger.info("Found trainer in default directory ('" + updater.workDir + "')...");
			updater.doUpdate(updateFullVersionData);
		} else if (updater.canUpdateByFileChooser()) {
		  logger.info("Found trainer by user selection via file chooser ('" + updater.workDir + "')...");
			updater.doUpdate(updateFullVersionData);
		} else {
			updater.gui.appendMessage(TrainerTextResources.getString("UPDATER_ERROR"));
			updater.exitAction.setEnabled(true);
		}

	}
	
	private boolean canUpdateInArgDirectory(final String argDir) {
	  if (argDir == null || argDir.length() < 1) {
	    logger.info("No argDir passed in.");
	    return false;
	  }
	  logger.info("Checking argDir: " + argDir);
    this.workDir = new File(argDir);
    return canUpdateInWorkingDirectory();
  }

  private static void loadUpdater(final boolean updateFullVersionData) {
		final Updater updater = new Updater(true);

    logger.info("Updater version " + (Util.getVersion() == null ? "N/A" : Util.getVersion()));
    
		final SwingWorker<File, Void> worker = new SwingWorker<File, Void>() {
      @Override
      protected File doInBackground() throws Exception {
        try {
          final File updaterFile = updater.downloadUpdater(updateFullVersionData);
          final File updateFile = getUpdateFile();
          if (updateFile.exists()) {
            updateFile.delete();
          }
          return updaterFile;
        } catch (IOException e) {
          logger.error("Failed to download a new updater version", e);
        }
        return null;
      }
      @Override
      protected void done() {
        super.done();
        try {
          logger.info("Initialising auto update by updater.jar...");
          File updaterFile = get();
          if (updaterFile == null) {
            updaterFile = new File(new File(System.getProperty("user.dir")), AutoUpdater.AUTO_UPDATER_JAR_NAME);
          }
          Runtime.getRuntime().exec("java -jar " + updaterFile.getName(), null, updaterFile.getParentFile());
        } catch (IOException e) {
          logger.error("Failed to launch updater", e);
        } catch (InterruptedException e) {
          logger.error("Failed to launch updater", e);
        } catch (ExecutionException e) {
          logger.error("Failed to launch updater", e);
        }

        logger.info("Attempting to exit after simple updater.jar download.");
        updater.exit();
      }
    };
    
    final String message = TrainerTextResources.getString("AUTO_UPDATE_LOADING_UPDATER");
    final PleaseWaitDialog pleaseWait = new PleaseWaitDialog(updater.gui, message, worker);
    SwingUtilities.invokeLater(new Runnable() {
      public void run() {
        pleaseWait.setVisible(true);
      }
    });
	}
  
  private static boolean isFullVersionUpdate() {
    File jarFile;
    try {
      jarFile = new File(Updater.class.getProtectionDomain().getCodeSource().getLocation().toURI().getPath());
      JarInputStream jarStream = new JarInputStream(new FileInputStream(jarFile));
      Manifest mf = jarStream.getManifest();
      jarStream.close();
      final String fullVersionUpdate = mf.getMainAttributes().getValue("Full-version-update");
      return "true".equals(fullVersionUpdate);
    } catch (URISyntaxException | IOException e) {
      return false;
    }
  }
	
}
